"""
PC-BASIC - video_pygame.py
Graphical interface based on PyGame

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging
import ctypes

if False:
    # for detection by packagers
    import pygame

from ..compat import iteritems, unichr
from ..basic.base import signals
from ..basic.base import scancode
from ..basic.base import bytematrix
from ..basic.base.eascii import as_unicode as uea
from ..compat import WIN32, MACOS, PY2
from ..compat import set_dpi_aware
from .video import VideoPlugin
from .base import video_plugins, InitFailed, EnvironmentCache, NOKILL_MESSAGE
from . import clipboard
from . import window


# refresh cycle parameters
# number of cycles to change blink state
BLINK_CYCLES = 5
# ms duration of a blink
BLINK_TIME = 120
CYCLE_TIME = BLINK_TIME // BLINK_CYCLES

# blank icon
BLANK_ICON = ((0,) * 16) * 16


@video_plugins.register('pygame')
class VideoPygame(VideoPlugin):
    """Pygame-based graphical interface."""

    def __init__(
            self, input_queue, video_queue,
            caption=u'', icon=BLANK_ICON,
            scaling=None, dimensions=None, aspect_ratio=(4, 3), border_width=0, fullscreen=False,
            prevent_close=False, mouse_clipboard=True,
            **kwargs
        ):
        """Initialise pygame interface."""
        logging.warning('The `pygame` interface is deprecated, please use the `graphical` interface instead.')
        try:
            _import_pygame()
        except ImportError:
            raise InitFailed('Module `pygame` not found')
        VideoPlugin.__init__(self, input_queue, video_queue)
        # request smooth scaling
        self._smooth = scaling == 'smooth'
        # ignore ALT+F4 and window X button
        self._nokill = prevent_close
        # window caption/title
        self.caption = caption
        # start in fullscreen mode
        self.fullscreen = fullscreen
        self._has_window = False
        # set state objects to whatever is now in state (may have been unpickled)
        # Windows 10 - set to DPI aware to avoid scaling twice on HiDPI screens
        set_dpi_aware()
        # ensure we have the correct video driver for SDL 1.2
        # pygame sets this on import, but if we've tried SDL2 we've had to
        # reset this value
        # ensure window is centred
        self._env = EnvironmentCache()
        self._env.set('SDL_VIDEO_CENTERED', '1')
        pygame.init()
        try:
            # poll the driver to force an exception if not initialised
            pygame.display.get_driver()
        except pygame.error as e:
            self._close_pygame()
            raise InitFailed('No suitable display driver: %s' % e)
        # display & border
        # display buffer
        self.canvas = None
        # border attribute
        self.border_attr = 0
        # palette and colours
        # working palette - attribute index in blue channel
        self.work_palette = [(0, 0, index) for index in range(256)]
        # display palettes for blink states 0, 1
        self._palette = [None, None]
        # composite colour artifacts
        self._pixel_packing = False
        # text attributes supported
        self._palette_blinks = True
        # update cycle
        # update flag
        self.busy = False
        # refresh cycle parameters
        self._cycle = 0
        self.last_cycle = 0
        # cursor
        # cursor shape
        self.cursor = None
        # current cursor location
        self.last_row = 1
        self.last_col = 1
        # cursor is visible
        self.cursor_visible = True
        self.cursor_attr = 7
        # buffer for text under cursor
        self.under_top_left = None
        # fonts
        # joystick and mouse
        # available joysticks
        self.joysticks = []
        # get physical screen dimensions (needs to be called before set_mode)
        display_info = pygame.display.Info()
        self._window_sizer = window.WindowSizer(
            display_info.current_w, display_info.current_h,
            scaling, dimensions, aspect_ratio, border_width,
        )
        # determine initial display size
        self._window_sizer.set_canvas_size(720, 400)
        self._set_icon(icon)
        try:
            self._resize_display()
        except pygame.error as e:
            self._close_pygame()
            raise InitFailed('Could not initialise display: %s' % e)
        if self._smooth and self.display.get_bitsize() < 24:
            logging.warning(
                'Smooth scaling not available on this display (depth %d < 24)',
                self.display.get_bitsize()
            )
            self._smooth = False
        if PY2: # pragma: no cover
            pygame.display.set_caption(self.caption.encode('utf-8', 'replace'))
        else:
            pygame.display.set_caption(self.caption)
        pygame.key.set_repeat(500, 24)
        # load an all-black 16-colour game palette to get started
        self.set_palette([((0,0,0), (0,0,0), False, False)]*16, None)
        pygame.joystick.init()
        self.joysticks = [pygame.joystick.Joystick(x) for x in range(pygame.joystick.get_count())]
        for j in self.joysticks:
            j.init()
        # if a joystick is present, its axes report 128 for mid, not 0
        for joy in range(len(self.joysticks)):
            for axis in (0, 1):
                self._input_queue.put(signals.Event(signals.STICK_MOVED, (joy, axis, 128)))
        # mouse setups
        self._mouse_clip = mouse_clipboard
        self.cursor_row, self.cursor_col = 1, 1
        # set_mode should be first event on queue
        self.f11_active = False
        self.clipboard_handler = get_clipboard_handler()
        # buffer for perceived alt key status, for use by workarounds
        self._alt_key_down = None

    def __exit__(self, type, value, traceback):
        """Close the pygame interface."""
        VideoPlugin.__exit__(self, type, value, traceback)
        self._close_pygame()

    def _close_pygame(self):
        """Close pygame modules and displays."""
        # if pygame import failed, close() is called while pygame is None
        if pygame:
            pygame.joystick.quit()
            pygame.display.quit()
            pygame.quit()
        # put environment variables back as they were
        self._env.close()

    def _set_icon(self, mask):
        """Set the window icon."""
        height, width = len(mask), len(mask[0])
        icon = pygame.Surface((width, height), depth=8) # pylint: disable=E1121,E1123
        icon.fill(0)
        array = bytematrix.ByteMatrix(height, width, mask)
        #icon.blit(glyph_to_surface(mask), (0, 0, width, height))
        pygame.surfarray.blit_array(icon, _BufferWrapper(array))
        icon.set_palette_at(0, (0, 0, 0))
        icon.set_palette_at(1, (0xff, 0xff, 0xff))
        pygame.transform.scale2x(icon)
        pygame.transform.scale2x(icon)
        pygame.display.set_icon(icon)


    ###########################################################################
    # input cycle

    def _check_input(self):
        """Handle screen and interface events."""
        # check and handle pygame events
        if not self._has_window:
            return
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                self._handle_key_down(event)
            elif event.type == pygame.KEYUP:
                self._handle_key_up(event)
            elif event.type == pygame.MOUSEBUTTONDOWN:
                if self._mouse_clip:
                    if event.button == 1:
                        # LEFT button: copy
                        pos = self._window_sizer.normalise_pos(*event.pos)
                        self.clipboard.start(
                            1 + pos[1] // self.font_height,
                            1 + (pos[0]+self.font_width//2) // self.font_width
                        )
                    elif event.button == 2:
                        # MIDDLE button: paste
                        text = self.clipboard_handler.paste()
                        self.clipboard.paste(text)
                    self.busy = True
                if event.button == 1:
                    # right mouse button is a pen press
                    self._input_queue.put(signals.Event(
                        signals.PEN_DOWN, self._window_sizer.normalise_pos(*event.pos)
                    ))
            elif event.type == pygame.MOUSEBUTTONUP:
                self._input_queue.put(signals.Event(signals.PEN_UP))
                if self._mouse_clip and event.button == 1:
                    self.clipboard.copy()
                    self.clipboard.stop()
                    self.busy = True
            elif event.type == pygame.MOUSEMOTION:
                pos = self._window_sizer.normalise_pos(*event.pos)
                self._input_queue.put(signals.Event(signals.PEN_MOVED, pos))
                if self.clipboard.active():
                    self.clipboard.move(
                        1 + pos[1] // self.font_height,
                        1 + (pos[0]+self.font_width//2) // self.font_width
                    )
                    self.busy = True
            elif event.type == pygame.JOYBUTTONDOWN:
                self._input_queue.put(signals.Event(
                    signals.STICK_DOWN, (event.joy, event.button)
                ))
            elif event.type == pygame.JOYBUTTONUP:
                self._input_queue.put(signals.Event(
                    signals.STICK_UP, (event.joy, event.button)
                ))
            elif event.type == pygame.JOYAXISMOTION:
                self._input_queue.put(signals.Event(
                    signals.STICK_MOVED, (event.joy, event.axis, int(event.value*127 + 128))
                ))
            elif event.type == pygame.VIDEORESIZE:
                if not self.fullscreen:
                    self._window_sizer.set_display_size(event.w, event.h)
                    self._resize_display()
            elif event.type == pygame.QUIT:
                if self._nokill:
                    self.set_caption_message(NOKILL_MESSAGE)
                else:
                    self._input_queue.put(signals.Event(signals.QUIT))

    def _handle_key_down(self, e):
        """Handle key-down event."""
        # get scancode
        scan = KEY_TO_SCAN.get(e.key, None)
        # get modifiers
        mod = [s for m, s in iteritems(MOD_TO_SCAN) if e.mod & m]
        # compensate for missing l-alt down events (needed at least on Ubuntu Unity)
        if (e.mod & pygame.KMOD_LALT) and not self._alt_key_down:
            # insert an alt keydown event before the alt-modified keydown
            fake_event = pygame.event.Event(
                pygame.KEYDOWN, scancode=0, key=pygame.K_LALT, unicode=u'', mod=0
            )
            self._handle_key_down(fake_event)
        if e.key == pygame.K_LALT:
            self._alt_key_down = True
        # get eascii
        try:
            if e.mod & pygame.KMOD_LALT:
                c = ALT_KEY_TO_EASCII[e.key]
            elif e.mod & pygame.KMOD_CTRL:
                c = CTRL_KEY_TO_EASCII[e.key]
            elif e.mod & pygame.KMOD_SHIFT:
                c = SHIFT_KEY_TO_EASCII[e.key]
            else:
                c = KEY_TO_EASCII[e.key]
        except KeyError:
            key = e.key
            if e.mod & pygame.KMOD_CTRL and key >= ord('a') and key <= ord('z'):
                c = unichr(key - ord('a') + 1)
            elif e.mod & pygame.KMOD_CTRL and key >= ord('[') and key <= ord('_'):
                c = unichr(key - ord('A') + 1)
            else:
                c = e.unicode
        # handle F11 home-key
        if e.key == pygame.K_F11:
            self.f11_active = True
            self.clipboard.start(self.cursor_row, self.cursor_col)
        elif self.f11_active:
            # F11+f to toggle fullscreen mode
            if e.key == pygame.K_f:
                self.fullscreen = not self.fullscreen
                self._window_sizer.set_canvas_size(*self.size, fullscreen=self.fullscreen)
                self._resize_display()
            self.clipboard.handle_key(scan, c)
            self.busy = True
        else:
            # double NUL characters, as single NUL signals e-ASCII
            if c == u'\0':
                c = uea.NUL
            # fix missing ascii for numeric keypad on Windows
            if e.mod & pygame.KMOD_NUM:
                try:
                    c = KEY_TO_EASCII_NUM[e.key]
                except KeyError:
                    pass
            # insert into keyboard queue
            self._input_queue.put(signals.Event(signals.KEYB_DOWN, (c, scan, mod)))

    def _handle_key_up(self, e):
        """Handle key-up event."""
        # compensate for missing l-alt down events (needed at least on Ubuntu Unity)
        if (e.key == pygame.K_LALT or e.mod & pygame.KMOD_LALT) and not self._alt_key_down:
            # insert an alt keydown event before the alt-modified keydown
            fake_event = pygame.event.Event(
                pygame.KEYDOWN, scancode=0, key=pygame.K_LALT, unicode=u'', mod=0
            )
            self._handle_key_down(fake_event)
            if e.key != pygame.K_LALT:
                # insert a keydown before the keyup
                if not hasattr(e, 'unicode'):
                    e.unicode = u''
                self._handle_key_down(e)
        if e.key == pygame.K_LALT:
            self._alt_key_down = False
        elif e.key == pygame.K_F11:
            self.clipboard.stop()
            self.busy = True
            self.f11_active = False
        # last key released gets remembered
        try:
            self._input_queue.put(signals.Event(signals.KEYB_UP, (KEY_TO_SCAN[e.key],)))
        except KeyError:
            pass


    ###########################################################################
    # screen drawing cycle

    def sleep(self, ms):
        """Sleep a tick to avoid hogging the cpu."""
        pygame.time.wait(ms)

    def _work(self):
        """Check screen and blink events; update screen if necessary."""
        if not self._has_window:
            return
        self.blink_state = 0
        if self._palette_blinks or self.text_cursor:
            self.blink_state = 0 if self._cycle < BLINK_CYCLES * 2 else 1
            if self._cycle % BLINK_CYCLES == 0:
                self.busy = True
        if self.cursor_visible and (
                (self.cursor_row != self.last_row) or (self.cursor_col != self.last_col)
            ):
            self.busy = True
        tock = pygame.time.get_ticks()
        if tock - self.last_cycle >= CYCLE_TIME:
            self.last_cycle = tock
            self._cycle += 1
            if self._cycle == BLINK_CYCLES * 4:
                self._cycle = 0
        if self.busy:
            self._do_flip()
            self.busy = False

    def _do_flip(self):
        """Draw the canvas to the screen."""
        # create the screen that will be stretched onto the display
        border_x, border_y = self._window_sizer.border_shift
        # surface depth and flags match those of canvas
        # pylint: disable=E1121,E1123
        screen = pygame.Surface(
            (self.size[0] + 2*border_x, self.size[1] + 2*border_y),
            0, self.canvas
        )
        screen.set_palette(self.work_palette)
        # border colour
        border_colour = pygame.Color(0, 0, self.border_attr % self.num_fore_attrs)
        screen.fill(border_colour)
        screen.blit(self.canvas, (border_x, border_y))
        # subsurface referencing the canvas area
        workscreen = screen.subsurface((border_x, border_y, self.size[0], self.size[1]))
        self._draw_cursor(workscreen)
        if self.clipboard.active():
            create_feedback(workscreen, self.clipboard.selection_rect)
        if self._pixel_packing:
            screen = apply_composite_artifacts(screen, *self._pixel_packing)
        screen.set_palette(self._palette[self.blink_state])
        letterbox = pygame.Rect(
            self._window_sizer.letterbox_shift, self._window_sizer.window_size
        )
        if self._smooth:
            # smoothscale to subsurface does not work correctly, so create a new surface and blit
            dest_surf = pygame.Surface(self._window_sizer.window_size) #, self.display)
            pygame.transform.smoothscale(
                screen.convert(dest_surf), self._window_sizer.window_size, dest_surf
            )
            self.display.blit(dest_surf, letterbox)
        else:
            dest_surf = self.display.subsurface(letterbox)
            pygame.transform.scale(
                screen.convert(dest_surf), self._window_sizer.window_size, dest_surf
            )
        pygame.display.flip()

    def _draw_cursor(self, screen):
        """Draw the cursor on the surface provided."""
        if not self.cursor_visible:
            return
        # copy screen under cursor
        self.under_top_left = (
            (self.cursor_col-1) * self.font_width,
            (self.cursor_row-1) * self.font_height
        )
        under_char_area = pygame.Rect(
            (self.cursor_col-1) * self.font_width, (self.cursor_row-1) * self.font_height,
            self.cursor_width, self.font_height
        )
        if self.text_cursor:
            # cursor is visible - to be done every cycle between 5 and 10, 15 and 20
            if self._cycle // BLINK_CYCLES in (1, 3):
                screen.blit(
                    self.cursor,
                    ((self.cursor_col-1) * self.font_width, (self.cursor_row-1) * self.font_height)
                )
        else:
            index = self.cursor_attr % self.num_fore_attrs
            # reference the destination area
            dest_array = pygame.surfarray.pixels2d(
                screen.subsurface(pygame.Rect(
                    (self.cursor_col-1) * self.font_width,
                    (self.cursor_row-1) * self.font_height + self.cursor_from,
                    self.cursor_width,
                    self.cursor_to - self.cursor_from + 1
                ))
            )
            dest_array ^= index
        self.last_row = self.cursor_row
        self.last_col = self.cursor_col

    ###########################################################################
    # miscellaneous helper functions

    def _resize_display(self):
        """Change the display size."""
        if self.fullscreen:
            flags = pygame.NOFRAME
        else:
            flags = pygame.RESIZABLE
        self.display = pygame.display.set_mode(self._window_sizer.display_size, flags)
        # load display if requested
        self.busy = True


    ###########################################################################
    # signal handlers

    def set_mode(self, canvas_height, canvas_width, text_height, text_width):
        """Initialise a given text or graphics mode."""
        # set display geometry
        self.font_height = -(-canvas_height // text_height)
        self.font_width = canvas_width // text_width
        # logical size
        self.size = canvas_width, canvas_height
        self._window_sizer.set_canvas_size(*self.size, fullscreen=self.fullscreen)
        self._resize_display()
        # set standard cursor
        self.cursor_width = self.font_width
        self.set_cursor_shape(0, self.font_height)
        # whole screen (blink on & off)
        self.canvas = pygame.Surface(self.size, depth=8) # pylint: disable=E1121,E1123
        self.canvas.set_palette(self.work_palette)
        # initialise clipboard
        self.clipboard = clipboard.ClipboardInterface(
            self.clipboard_handler, self._input_queue,
            text_width, text_height, self.font_width, self.font_height, self.size
        )
        self.busy = True
        self._has_window = True

    def set_caption_message(self, msg):
        """Add a message to the window caption."""
        title = self.caption + (u' - ' + msg if msg else u'')
        pygame.display.set_caption(title.encode('utf-8', 'replace'))

    def set_clipboard_text(self, text):
        """Put text on the clipboard."""
        self.clipboard_handler.copy(text)

    def set_palette(self, attributes, pack_pixels):
        """Build the palette."""
        self.num_fore_attrs = 16
        self.num_back_attrs = 8
        # fill up the 8-bit palette with all combinations we need
        # blink states: 0 light up, 1 light down
        # bottom 128 are non-blink, top 128 blink to background
        self._palette[0] = [_fore for _fore, _, _, _ in attributes]
        self._palette[1] = [_back if _blink else _fore for _fore, _back, _blink, _ in attributes]
        self._palette_blinks = self._palette[0] != self._palette[1]
        self._pixel_packing = pack_pixels
        self.busy = True

    def set_border_attr(self, attr):
        """Change the border attribute."""
        self.border_attr = attr
        self.busy = True

    def clear_rows(self, back_attr, start, stop):
        """Clear a range of screen rows."""
        bg = (0, 0, back_attr)
        scroll_area = pygame.Rect(
            0, (start-1)*self.font_height, self.size[0], (stop-start+1)*self.font_height
        )
        self.canvas.fill(bg, scroll_area)
        self.busy = True

    def show_cursor(self, cursor_on, cursor_blinks):
        """Change visibility of cursor."""
        self.cursor_visible = cursor_on
        self.text_cursor = cursor_blinks
        self.busy = True

    def move_cursor(self, row, col, attr, width):
        """Move the cursor to a new position."""
        self.cursor_row, self.cursor_col = row, col
        # set attribute
        self.cursor_attr = attr % self.num_fore_attrs
        # set width
        if width != self.cursor_width:
            self.cursor_width = width
            self._rebuild_cursor()
        else:
            self.cursor.set_palette_at(254, pygame.Color(0, self.cursor_attr, self.cursor_attr))

    def scroll(self, direction, from_line, scroll_height, back_attr):
        """Scroll the screen between from_line and scroll_height."""
        temp_scroll_area = pygame.Rect(
            0, (from_line-1)*self.font_height,
            self.size[0], (scroll_height-from_line+1) * self.font_height
        )
        # scroll
        self.canvas.set_clip(temp_scroll_area)
        self.canvas.scroll(0, direction * self.font_height)
        # empty new line
        bg = (0, 0, back_attr)
        self.canvas.fill(
            bg, (0, (scroll_height-1) * self.font_height, self.size[0], self.font_height)
        )
        self.canvas.set_clip(None)
        self.busy = True

    def set_cursor_shape(self, from_line, to_line):
        """Build a sprite for the cursor."""
        self.cursor_from, self.cursor_to = from_line, to_line
        self._rebuild_cursor()

    def _rebuild_cursor(self):
        """Rebuild cursor surface."""
        height = self.font_height
        width = self.cursor_width
        from_line, to_line = self.cursor_from, self.cursor_to
        self.cursor = pygame.Surface((width, height), depth=8) # pylint: disable=E1121,E1123
        color, bg = 254, 255
        self.cursor.set_colorkey(bg)
        self.cursor.fill(bg)
        self.cursor.fill(color, (0, from_line, width, min(to_line-from_line+1, height-from_line)))
        self.cursor.set_palette_at(254, pygame.Color(0, self.cursor_attr, self.cursor_attr))
        self.busy = True

    def update(self, row, col, unicode_matrix, attr_matrix, y0, x0, array):
        """Put text or pixels at a given position."""
        if y0 + array.height > self.size[1] or x0 +array. width > self.size[0]:
            array = array[:self.size[1]-y0, :self.size[0]-x0]
        # reference the destination area
        subsurface = self.canvas.subsurface(pygame.Rect(x0, y0, array.width, array.height))
        pygame.surfarray.blit_array(subsurface, _BufferWrapper(array))
        self.busy = True


class _BufferWrapper(object):

    def __init__(self, array):
        self._buffer = ctypes.create_string_buffer(array.to_bytes())
        self.__array_interface__ = dict(
            shape=(array.width, array.height),
            strides=(1, array.width),
            typestr='|u1', version=3, data=(ctypes.addressof(self._buffer), False)
        )



###############################################################################
# clipboard handling

class PygameClipboard(clipboard.Clipboard):
    """Clipboard handling using Pygame.Scrap."""

    # text type we look for in the clipboard
    text = (
        'UTF8_STRING', 'text/plain;charset=utf-8', 'text/plain',
        'TEXT', 'STRING'
    )

    def __init__(self):
        """Initialise the clipboard handler."""
        try:
            pygame.scrap.init()
            self.ok = True
        except Exception:
            if pygame:
                logging.warning('PyGame.Scrap clipboard handling module not found.')
            self.ok = False

    def copy(self, text):
        """Put unicode text on clipboard."""
        pygame.scrap.set_mode(pygame.SCRAP_CLIPBOARD)
        try:
            if WIN32:
                # on Windows, encode as utf-16 without FF FE byte order mark and null-terminate
                # but give it a utf-8 MIME type, because that's how Windows likes it
                pygame.scrap.put(
                    'text/plain;charset=utf-8', text.encode('utf-16le', 'replace') + b'\0\0'
                )
            else:
                pygame.scrap.put(pygame.SCRAP_TEXT, text.encode('utf-8', 'replace'))
        except Exception as e:# KeyError:
            logging.debug('Clipboard copy failed for clip %r: %s', text, e)

    def paste(self):
        """Return unicode text from clipboard."""
        pygame.scrap.set_mode(pygame.SCRAP_CLIPBOARD)
        us = u''
        s = b''
        available = pygame.scrap.get_types()
        for text_type in self.text:
            if text_type not in available:
                continue
            s = pygame.scrap.get(text_type)
            if s:
                break
        if WIN32:
            if text_type == 'text/plain;charset=utf-8':
                # it's lying, it's giving us UTF16 little-endian
                # ignore any bad UTF16 characters from outside
                us = s.decode('utf-16le', errors='replace')
            else:
                # fallback
                us = s.decode('ascii', errors='replace')
            # remove null-terminator
            us = us[:us.find(u'\0')]
        else:
            us = s.decode('utf-8', errors='replace')
        if us:
            us = us.replace(u'\r\n', u'\n').replace(u'\n', u'\r')
        return us or u''

def get_clipboard_handler():
    """Get a working Clipboard handler object."""
    # Pygame.Scrap doesn't work on OSX and crashes on Linux
    if MACOS:
        handler = clipboard.MacClipboard()
    elif WIN32:
        handler = PygameClipboard()
    else:
        handler = clipboard.XClipboard()
    if not handler.ok:
        logging.warning('Clipboard copy and paste not available.')
        handler = clipboard.Clipboard()
    return handler

def create_feedback(surface, selection_rects):
    """Create visual feedback for selection onto a surface."""
    for r in selection_rects:
        work_area = surface.subsurface(pygame.Rect(r))
        orig = work_area.copy()
        # add 1 to the color as a highlight
        orig.fill(pygame.Color(0, 0, 1))
        work_area.blit(orig, (0, 0), special_flags=pygame.BLEND_ADD)


###############################################################################
# import PyGame and define constants

pygame = None


def _import_pygame():
    """Import pygame and define constants."""
    global pygame
    global KEY_TO_SCAN, MOD_TO_SCAN, KEY_TO_EASCII_NUM
    global KEY_TO_EASCII, SHIFT_KEY_TO_EASCII, CTRL_KEY_TO_EASCII, ALT_KEY_TO_EASCII

    import pygame

    # these are PC keyboard scancodes
    KEY_TO_SCAN = {
        # top row
        pygame.K_ESCAPE: scancode.ESCAPE, pygame.K_1: scancode.N1,
        pygame.K_2: scancode.N2, pygame.K_3: scancode.N3,
        pygame.K_4: scancode.N4, pygame.K_5: scancode.N5,
        pygame.K_6: scancode.N6, pygame.K_7: scancode.N7,
        pygame.K_8: scancode.N8, pygame.K_9: scancode.N9,
        pygame.K_0: scancode.N0, pygame.K_MINUS: scancode.MINUS,
        pygame.K_EQUALS: scancode.EQUALS,
        pygame.K_BACKSPACE: scancode.BACKSPACE,
        # row 1
        pygame.K_TAB: scancode.TAB, pygame.K_q: scancode.q,
        pygame.K_w: scancode.w, pygame.K_e: scancode.e, pygame.K_r: scancode.r,
        pygame.K_t: scancode.t, pygame.K_y: scancode.y, pygame.K_u: scancode.u,
        pygame.K_i: scancode.i, pygame.K_o: scancode.o, pygame.K_p: scancode.p,
        pygame.K_LEFTBRACKET: scancode.LEFTBRACKET,
        pygame.K_RIGHTBRACKET: scancode.RIGHTBRACKET,
        pygame.K_RETURN: scancode.RETURN, pygame.K_KP_ENTER: scancode.RETURN,
        # row 2
        pygame.K_RCTRL: scancode.CTRL, pygame.K_LCTRL: scancode.CTRL,
        pygame.K_a: scancode.a, pygame.K_s: scancode.s, pygame.K_d: scancode.d,
        pygame.K_f: scancode.f, pygame.K_g: scancode.g, pygame.K_h: scancode.h,
        pygame.K_j: scancode.j, pygame.K_k: scancode.k, pygame.K_l: scancode.l,
        pygame.K_SEMICOLON: scancode.SEMICOLON, pygame.K_QUOTE: scancode.QUOTE,
        pygame.K_BACKQUOTE: scancode.BACKQUOTE,
        # row 3
        pygame.K_LSHIFT: scancode.LSHIFT,
        pygame.K_BACKSLASH: scancode.BACKSLASH,
        pygame.K_z: scancode.z, pygame.K_x: scancode.x, pygame.K_c: scancode.c,
        pygame.K_v: scancode.v, pygame.K_b: scancode.b, pygame.K_n: scancode.n,
        pygame.K_m: scancode.m, pygame.K_COMMA: scancode.COMMA,
        pygame.K_PERIOD: scancode.PERIOD, pygame.K_SLASH: scancode.SLASH,
        pygame.K_RSHIFT: scancode.RSHIFT, pygame.K_PRINT: scancode.PRINT,
        pygame.K_SYSREQ: scancode.SYSREQ,
        pygame.K_LALT: scancode.ALT,
        # don't catch right-Alt as it may inhibit AltGr on Windows
        #pygame.K_RALT: scancode.ALT,
        pygame.K_SPACE: scancode.SPACE, pygame.K_CAPSLOCK: scancode.CAPSLOCK,
        # function key row
        pygame.K_F1: scancode.F1, pygame.K_F2: scancode.F2,
        pygame.K_F3: scancode.F3, pygame.K_F4: scancode.F4,
        pygame.K_F5: scancode.F5, pygame.K_F6: scancode.F6,
        pygame.K_F7: scancode.F7, pygame.K_F8: scancode.F8,
        pygame.K_F9: scancode.F9, pygame.K_F10: scancode.F10,
        pygame.K_F11: scancode.F11, pygame.K_F12: scancode.F12,
        # top middle
        pygame.K_NUMLOCK: scancode.NUMLOCK,
        pygame.K_SCROLLOCK: scancode.SCROLLOCK,
        # keypad
        pygame.K_KP7: scancode.KP7, pygame.K_HOME: scancode.HOME,
        pygame.K_KP8: scancode.KP8, pygame.K_UP: scancode.UP,
        pygame.K_KP9: scancode.KP9, pygame.K_PAGEUP: scancode.PAGEUP,
        pygame.K_KP_MINUS: scancode.KPMINUS,
        pygame.K_KP4: scancode.KP4, pygame.K_LEFT: scancode.LEFT,
        pygame.K_KP5: scancode.KP5,
        pygame.K_KP6: scancode.KP6, pygame.K_RIGHT: scancode.RIGHT,
        pygame.K_KP_PLUS: scancode.KPPLUS,
        pygame.K_KP1: scancode.KP1, pygame.K_END: scancode.END,
        pygame.K_KP2: scancode.KP2, pygame.K_DOWN: scancode.DOWN,
        pygame.K_KP3: scancode.KP3, pygame.K_PAGEDOWN: scancode.PAGEDOWN,
        pygame.K_KP0: scancode.KP0, pygame.K_INSERT: scancode.INSERT,
        # keypad dot, times, div, enter ?
        # various
        pygame.K_DELETE: scancode.DELETE,
        pygame.K_PAUSE: scancode.BREAK,
        pygame.K_BREAK: scancode.BREAK,
    }

    KEY_TO_EASCII = {
        pygame.K_F1: uea.F1,
        pygame.K_F2: uea.F2,
        pygame.K_F3: uea.F3,
        pygame.K_F4: uea.F4,
        pygame.K_F5: uea.F5,
        pygame.K_F6: uea.F6,
        pygame.K_F7: uea.F7,
        pygame.K_F8: uea.F8,
        pygame.K_F9: uea.F9,
        pygame.K_F10: uea.F10,
        pygame.K_F11: uea.F11,
        pygame.K_F12: uea.F12,
        pygame.K_HOME: uea.HOME,
        pygame.K_UP: uea.UP,
        pygame.K_PAGEUP: uea.PAGEUP,
        pygame.K_LEFT: uea.LEFT,
        pygame.K_RIGHT: uea.RIGHT,
        pygame.K_END: uea.END,
        pygame.K_DOWN: uea.DOWN,
        pygame.K_PAGEDOWN: uea.PAGEDOWN,
        pygame.K_ESCAPE: uea.ESCAPE,
        pygame.K_BACKSPACE: uea.BACKSPACE,
        pygame.K_TAB: uea.TAB,
        pygame.K_RETURN: uea.RETURN,
        pygame.K_KP_ENTER: uea.RETURN,
        pygame.K_SPACE: uea.SPACE,
        pygame.K_INSERT: uea.INSERT,
        pygame.K_DELETE: uea.DELETE,
    }

    SHIFT_KEY_TO_EASCII = {
        pygame.K_F1: uea.SHIFT_F1,
        pygame.K_F2: uea.SHIFT_F2,
        pygame.K_F3: uea.SHIFT_F3,
        pygame.K_F4: uea.SHIFT_F4,
        pygame.K_F5: uea.SHIFT_F5,
        pygame.K_F6: uea.SHIFT_F6,
        pygame.K_F7: uea.SHIFT_F7,
        pygame.K_F8: uea.SHIFT_F8,
        pygame.K_F9: uea.SHIFT_F9,
        pygame.K_F10: uea.SHIFT_F10,
        pygame.K_F11: uea.SHIFT_F11,
        pygame.K_F12: uea.SHIFT_F12,
        pygame.K_HOME: uea.SHIFT_HOME,
        pygame.K_UP: uea.SHIFT_UP,
        pygame.K_PAGEUP: uea.SHIFT_PAGEUP,
        pygame.K_LEFT: uea.SHIFT_LEFT,
        pygame.K_RIGHT: uea.SHIFT_RIGHT,
        pygame.K_END: uea.SHIFT_END,
        pygame.K_DOWN: uea.SHIFT_DOWN,
        pygame.K_PAGEDOWN: uea.SHIFT_PAGEDOWN,
        pygame.K_ESCAPE: uea.SHIFT_ESCAPE,
        pygame.K_BACKSPACE: uea.SHIFT_BACKSPACE,
        pygame.K_TAB: uea.SHIFT_TAB,
        pygame.K_RETURN: uea.SHIFT_RETURN,
        pygame.K_KP_ENTER: uea.SHIFT_RETURN,
        pygame.K_SPACE: uea.SHIFT_SPACE,
        pygame.K_INSERT: uea.SHIFT_INSERT,
        pygame.K_DELETE: uea.SHIFT_DELETE,
        pygame.K_KP5: uea.SHIFT_KP5,
    }

    CTRL_KEY_TO_EASCII = {
        pygame.K_F1: uea.CTRL_F1,
        pygame.K_F2: uea.CTRL_F2,
        pygame.K_F3: uea.CTRL_F3,
        pygame.K_F4: uea.CTRL_F4,
        pygame.K_F5: uea.CTRL_F5,
        pygame.K_F6: uea.CTRL_F6,
        pygame.K_F7: uea.CTRL_F7,
        pygame.K_F8: uea.CTRL_F8,
        pygame.K_F9: uea.CTRL_F9,
        pygame.K_F10: uea.CTRL_F10,
        pygame.K_F11: uea.CTRL_F11,
        pygame.K_F12: uea.CTRL_F12,
        pygame.K_HOME: uea.CTRL_HOME,
        pygame.K_PAGEUP: uea.CTRL_PAGEUP,
        pygame.K_LEFT: uea.CTRL_LEFT,
        pygame.K_RIGHT: uea.CTRL_RIGHT,
        pygame.K_END: uea.CTRL_END,
        pygame.K_PAGEDOWN: uea.CTRL_PAGEDOWN,
        pygame.K_ESCAPE: uea.CTRL_ESCAPE,
        pygame.K_BACKSPACE: uea.CTRL_BACKSPACE,
        pygame.K_TAB: uea.CTRL_TAB,
        pygame.K_RETURN: uea.CTRL_RETURN,
        pygame.K_KP_ENTER: uea.CTRL_RETURN,
        pygame.K_SPACE: uea.CTRL_SPACE,
        pygame.K_PRINT: uea.CTRL_PRINT,
        pygame.K_2: uea.CTRL_2,
        pygame.K_6: uea.CTRL_6,
        pygame.K_MINUS: uea.CTRL_MINUS,
    }

    ALT_KEY_TO_EASCII = {
        pygame.K_1: uea.ALT_1,
        pygame.K_2: uea.ALT_2,
        pygame.K_3: uea.ALT_3,
        pygame.K_4: uea.ALT_4,
        pygame.K_5: uea.ALT_5,
        pygame.K_6: uea.ALT_6,
        pygame.K_7: uea.ALT_7,
        pygame.K_8: uea.ALT_8,
        pygame.K_9: uea.ALT_9,
        pygame.K_0: uea.ALT_0,
        pygame.K_MINUS: uea.ALT_MINUS,
        pygame.K_EQUALS: uea.ALT_EQUALS,
        pygame.K_q: uea.ALT_q,
        pygame.K_w: uea.ALT_w,
        pygame.K_e: uea.ALT_e,
        pygame.K_r: uea.ALT_r,
        pygame.K_t: uea.ALT_t,
        pygame.K_y: uea.ALT_y,
        pygame.K_u: uea.ALT_u,
        pygame.K_i: uea.ALT_i,
        pygame.K_o: uea.ALT_o,
        pygame.K_p: uea.ALT_p,
        pygame.K_a: uea.ALT_a,
        pygame.K_s: uea.ALT_s,
        pygame.K_d: uea.ALT_d,
        pygame.K_f: uea.ALT_f,
        pygame.K_g: uea.ALT_g,
        pygame.K_h: uea.ALT_h,
        pygame.K_j: uea.ALT_j,
        pygame.K_k: uea.ALT_k,
        pygame.K_l: uea.ALT_l,
        pygame.K_z: uea.ALT_z,
        pygame.K_x: uea.ALT_x,
        pygame.K_c: uea.ALT_c,
        pygame.K_v: uea.ALT_v,
        pygame.K_b: uea.ALT_b,
        pygame.K_n: uea.ALT_n,
        pygame.K_m: uea.ALT_m,
        pygame.K_F1: uea.ALT_F1,
        pygame.K_F2: uea.ALT_F2,
        pygame.K_F3: uea.ALT_F3,
        pygame.K_F4: uea.ALT_F4,
        pygame.K_F5: uea.ALT_F5,
        pygame.K_F6: uea.ALT_F6,
        pygame.K_F7: uea.ALT_F7,
        pygame.K_F8: uea.ALT_F8,
        pygame.K_F9: uea.ALT_F9,
        pygame.K_F10: uea.ALT_F10,
        pygame.K_F11: uea.ALT_F11,
        pygame.K_F12: uea.ALT_F12,
        pygame.K_BACKSPACE: uea.ALT_BACKSPACE,
        pygame.K_TAB: uea.ALT_TAB,
        pygame.K_RETURN: uea.ALT_RETURN,
        pygame.K_KP_ENTER: uea.ALT_RETURN,
        pygame.K_SPACE: uea.ALT_SPACE,
        pygame.K_PRINT: uea.ALT_PRINT,
        pygame.K_KP5: uea.ALT_KP5,
    }

    MOD_TO_SCAN = {
        pygame.KMOD_LSHIFT: scancode.LSHIFT,
        pygame.KMOD_RSHIFT: scancode.RSHIFT,
        pygame.KMOD_LCTRL: scancode.CTRL,
        pygame.KMOD_RCTRL: scancode.CTRL,
        pygame.KMOD_LALT: scancode.ALT,
        # don't catch right-Alt as it may inhibit AltGr on Windows
        #pygame.KMOD_RALT: scancode.ALT,
    }

    KEY_TO_EASCII_NUM = {
        pygame.K_KP0: u'0',
        pygame.K_KP1: u'1',
        pygame.K_KP2: u'2',
        pygame.K_KP3: u'3',
        pygame.K_KP4: u'4',
        pygame.K_KP5: u'5',
        pygame.K_KP6: u'6',
        pygame.K_KP7: u'7',
        pygame.K_KP8: u'8',
        pygame.K_KP9: u'9',
    }


def apply_composite_artifacts(screen, bpp_out, bpp_in):
    """Process the canvas to apply composite colour artifacts."""
    # not implemented
    return screen



"""
PC-BASIC - bytestream.py
BytesIO extension with externally provided buffer

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from _pyio import BytesIO, BufferedIOBase

from ...compat import PY2


class ByteStream(BytesIO):
    """BytesIO over external buffer."""

    def __init__(self, initial_bytes):
        """Create new ByteStream."""
        BytesIO.__init__(self)
        # use the actual object as a buffer, do not copy
        self._buffer = memoryview(initial_bytes)

    def write(self, b):
        if self._pos + len(b) > len(self._buffer):
            raise ValueError("Can't change size of buffer.")
        return BytesIO.write(self, b)

    def close(self):
        # _pyio.BytesIO clears buffer here but memoryview has no clear()
        BufferedIOBase.close(self)

    if PY2: # pragma: no cover
        def read(self, n=None):
            if self.closed:
                raise ValueError("read from closed file")
            if n is None:
                n = -1
            if not isinstance(n, (int, long)):
                raise TypeError("integer argument expected, got {0!r}".format(
                    type(n)))
            if n < 0:
                n = len(self._buffer)
            if len(self._buffer) <= self._pos:
                return b""
            newpos = min(len(self._buffer), self._pos + n)
            b = self._buffer[self._pos : newpos]
            self._pos = newpos
            # on py2, we have to convert to bytesarray before converting to bytes
            # or we'll get strings like '<memoryview at ...>'
            return bytes(bytearray(b))



"""
PC-BASIC - inputs package
Keyboard, pen and joystick handling

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from .keyboard import Keyboard
from .pen import Pen
from .stick import Stick



"""
PC-BASIC - eventcycle.py
Event queue handling

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import time

from ..compat import queue

from .base import error
from .base import scancode
from .base import signals
from .base.eascii import as_bytes as ea
from .base.eascii import as_unicode as uea


# F12 emulator home-key
# also f12+b -> ctrl+break
HOME_KEY_REPLACEMENTS_SCANCODE = {
    scancode.LEFT: (scancode.KP4, u'4'),
    scancode.RIGHT: (scancode.KP6, u'6'),
    scancode.UP: (scancode.KP8, u'8'),
    scancode.DOWN: (scancode.KP2, u'2'),
    # catch numbers by scancode, not eACSII
    # becasue the eASCII for Alt+number is different and that
    # will break inserting Alt+keypad numbers as Alt+F12+numbers
    scancode.N0: (scancode.KP0, u'0'),
    scancode.N1: (scancode.KP1, u'1'),
    scancode.N2: (scancode.KP2, u'2'),
    scancode.N3: (scancode.KP3, u'3'),
    scancode.N4: (scancode.KP4, u'4'),
    scancode.N5: (scancode.KP5, u'5'),
    scancode.N6: (scancode.KP6, u'6'),
    scancode.N7: (scancode.KP7, u'7'),
    scancode.N8: (scancode.KP8, u'8'),
    scancode.N9: (scancode.KP9, u'9'),
}

HOME_KEY_REPLACEMENTS_EASCII = {
    u'+': (scancode.KPPLUS, u'+'),
    u'-': (scancode.KPMINUS, u'-'),
    u'P': (scancode.BREAK, u''),
    u'N': (scancode.NUMLOCK, u''),
    u'S': (scancode.SCROLLOCK, u''),
    u'C': (scancode.CAPSLOCK, u''),
    u'H': (scancode.PRINT, u''),
    # ctrl+H
    u'\x08': (scancode.PRINT, uea.CTRL_PRINT),
}


###############################################################################
# queues

class NullQueue(object):
    """Dummy implementation of Queue interface."""
    def __init__(self, maxsize=0):
        pass
    def qsize(self):
        return 0
    def empty(self):
        return True
    def full(self):
        return False
    def put(self, item, block=False, timeout=False):
        pass
    def put_nowait(self, item):
        pass
    def get(self, block=False, timeout=False):
        # we're ignoring block
        raise queue.Empty
    def task_done(self):
        pass
    def join(self):
        pass


class EventQueues(object):
    """Manage interface queues."""

    tick = 0.006
    max_video_qsize = 200
    #max_audio_qsize = 20

    def __init__(self, ctrl_c_is_break, inputs=None, video=None, audio=None):
        """Initialise; default is NullQueues."""
        # input signal handlers
        self._handlers = []
        # basic event handlers
        self._basic_handlers = ()
        # pause-key halts everything until another keypress
        self._pause = False
        # treat ctrl+c as break interrupt
        self._ctrl_c_is_break = ctrl_c_is_break
        # F12 replacement events
        self._f12_active = False
        self.set(inputs, video, audio)

    def set(self, inputs=None, video=None, audio=None):
        """Set; default is NullQueues."""
        self.inputs = inputs or NullQueue()
        self.video = video or NullQueue()
        self.audio = audio or NullQueue()

    def __getstate__(self):
        """Don't pickle queues."""
        pickle_dict = self.__dict__.copy()
        pickle_dict['inputs'] = None
        pickle_dict['video'] = None
        pickle_dict['audio'] = None
        return pickle_dict

    def __setstate__(self, pickle_dict):
        """Set to null queues on unpickling."""
        self.__dict__.update(pickle_dict)
        self.set()

    def add_handler(self, handler):
        """Add an input handler."""
        self._handlers.append(handler)

    def set_basic_event_handlers(self, event_check_input):
        """Set the handlers for BASIC events."""
        self._basic_handlers = tuple(event_check_input)

    def wait(self):
        """Wait and check events."""
        time.sleep(self.tick)
        self.check_events()

    def check_events(self):
        """Main event cycle."""
        # sleep(0) is needed for responsiveness, e.g. event trapping in programs with tight loops
        # i.e. 100 goto 100 with event traps active) - needed to allow the input queue to fill
        # this also allows the screen to update between statements
        # it does slow the interpreter down by about 20% in FOR loops
        # note that we always have an input queue, either for the interface of for iostreams
        time.sleep(0)
        # bizarrely, we need sleep(0) twice. I don't know why.
        # note that multiple sleep(0) calls doen't seem to cause more slowdown than just one.
        time.sleep(0)
        time.sleep(0)
        # what I think is happening here is that the sdl2 interface thread,
        # in its loop to process a single queue item, calls C library functions
        # which do not need the GIL. so it releases it. this allows the engine thread to pick up
        # and produce more work for the interface thread. sleep(0) does not wait but it does release
        # the GIL back, so we need a few sleep(0) calls to allow the interface to get through
        # individual items.
        # this would not be a problem if the interface thread did not need the GIL at all
        # (perhaps with numba, nuitka, cython, pypy or jython)
        # but note that the video queue is a Python object so may require the GIL
        # or if it held the GIL for a full cycle
        # wait to for the queue to drain if it excceds a threshold value
        # this allows the interface to catch up with video updates
        if self.video.qsize() > self.max_video_qsize:
            while self.video.qsize():
                time.sleep(self.tick)
        self._check_input()

    def _check_input(self):
        """Handle input events."""
        while True:
            # pop input queues
            try:
                signal = self.inputs.get(False)
            except queue.Empty:
                if self._pause:
                    time.sleep(self.tick)
                    continue
                else:
                    # we still need to handle basic events: not all are inputs
                    for e in self._basic_handlers:
                        e.check_input(signals.Event(None))
                    break
            self.inputs.task_done()
            # effect replacements
            self._replace_inputs(signal)
            # handle input events
            for handle_input in (
                    [self._handle_non_trappable_interrupts] +
                    [e.check_input for e in self._basic_handlers] +
                    [self._handle_trappable_interrupts] +
                    [e.check_input for e in self._handlers]
                ):
                if handle_input(signal):
                    break

    def _handle_non_trappable_interrupts(self, signal):
        """Handle non-trappable interrupts (before BASIC events)."""
        # process input events
        if signal.event_type == signals.QUIT:
            raise error.Exit()
        # exit pause mode on keyboard hit; swallow key
        elif signal.event_type in (signals.KEYB_DOWN, signals.STREAM_CHAR, signals.CLIP_PASTE):
            if self._pause:
                self._pause = False
                return True
        return False

    def _handle_trappable_interrupts(self, signal):
        """Handle trappable interrupts (after BASIC events)."""
        # handle special key combinations
        if signal.event_type == signals.KEYB_DOWN:
            c, scan, mod = signal.params
            if (scan == scancode.DELETE and scancode.CTRL in mod and scancode.ALT in mod):
                # ctrl-alt-del: if not captured by the OS, reset the emulator
                # meaning exit and delete state. This is useful on android.
                raise error.Reset()
            elif scan in (scancode.BREAK, scancode.SCROLLOCK) and scancode.CTRL in mod:
                raise error.Break()
            # pause key handling
            # to ensure this key remains trappable
            elif (scan == scancode.BREAK or
                    (scan == scancode.NUMLOCK and scancode.CTRL in mod)):
                self._pause = True
                return True
        return False

    def _replace_inputs(self, signal):
        """Input event replacements."""
        if signal.event_type == signals.KEYB_DOWN:
            c, scan, mod = signal.params
            if (self._ctrl_c_is_break and c == uea.CTRL_c):
                # replace ctrl+c with ctrl+break if option is enabled
                signal.params = u'', scancode.BREAK, [scancode.CTRL]
            elif scan == scancode.F12:
                # F12 emulator "home key"
                self._f12_active = True
                signal.event_type = None
            elif self._f12_active:
                # F12 replacements
                if c.upper() == u'B':
                    # f12+b -> ctrl+break
                    signal.params = u'', scancode.BREAK, [scancode.CTRL]
                else:
                    scan, c = HOME_KEY_REPLACEMENTS_SCANCODE.get(scan, (scan, c))
                    scan, c = HOME_KEY_REPLACEMENTS_EASCII.get(c.upper(), (scan, c))
                    signal.params = c, scan, mod
        elif (signal.event_type == signals.KEYB_UP) and (signal.params[0] == scancode.F12):
            self._f12_active = False



"""
PC-BASIC - compat.win32_console
Windows console support:
- unicode output for Python 2
- scroll prevention
- ANSI input & adjustable echo

(c) 2018--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.

Borrows code from colorama, which is Copyright Jonathan Hartley 2013. BSD 3-Clause license
"""

# pylint: disable=no-name-in-module, no-member

import os
import sys
import time
import msvcrt
import ctypes
from contextlib import contextmanager
from collections import deque
from ctypes import windll, wintypes, POINTER, byref, Structure, cast

from .base import PY2
from .streams import StdIOBase, StreamWrapper

if PY2: # pragma: no cover
    from .python2 import SimpleNamespace
else:
    from types import SimpleNamespace



# Windows virtual key codes, mapped to standard key names
KEYS = SimpleNamespace(
    PAGEUP = 0x21, # VK_PRIOR
    PAGEDOWN = 0x22, # VK_NEXT
    END = 0x23,
    HOME = 0x24,
    LEFT = 0x25,
    UP = 0x26,
    RIGHT = 0x27,
    DOWN = 0x28,
    INSERT = 0x2d,
    DELETE = 0x2e,
    F1 = 0x70,
    F2 = 0x71,
    F3 = 0x72,
    F4 = 0x73,
    F5 = 0x74,
    F6 = 0x75,
    F7 = 0x76,
    F8 = 0x77,
    F9 = 0x78,
    F10 = 0x79,
    F11 = 0x7a,
    F12 = 0x7b,
    #
    ALT = 0x12, # VK_MENU
)
VK_TO_KEY = {value: key for key, value in KEYS.__dict__.items()}
# alpha key codes
VK_TO_KEY.update({
    value: chr(value).lower() for value in range(0x30, 0x5b)
})

# control key state bit flags
# CAPSLOCK_ON = 0x0080,
# ENHANCED_KEY = 0x0100,
# LEFT_ALT_PRESSED = 0x0002,
# LEFT_CTRL_PRESSED = 0x0008,
# NUMLOCK_ON = 0x0020,
# RIGHT_ALT_PRESSED = 0x0001,
# RIGHT_CTRL_PRESSED = 0x0004,
# SCROLLLOCK_ON = 0x0040,
# SHIFT_PRESSED = 0x0010,
MODS = dict(
    CTRL = 0x0c,
    ALT = 0x03,
    SHIFT = 0x10,
)

# character attributes, from wincon.h
NORMAL = 0x00
BRIGHT = 0x08


##############################################################################
# ctypes wrappers

class KEY_EVENT_RECORD(Structure):
    _fields_ = (
        ('bKeyDown', wintypes.BOOL), #32 bit?
        ('wRepeatCount', wintypes.WORD), #16
        ('wVirtualKeyCode', wintypes.WORD),#16
        ('wVirtualScanCode', wintypes.WORD), #16
        # union with CHAR AsciiChar
        ('UnicodeChar', wintypes.WCHAR), #32
        ('dwControlKeyState', wintypes.DWORD), #32
        # note that structure is in a union with other event records
        # but it is the largest type. mouseeventrecord is 128 bytes
    )

class INPUT_RECORD(Structure):
    _fields_ = (
        ('EventType', wintypes.WORD),
        # union of many event types but we only care about key events
        # total size is 16 bytes
        ('KeyEvent', KEY_EVENT_RECORD),
    )

class CHAR_INFO(Structure):
    _fields_ = (
        ('UnicodeChar', wintypes.WCHAR),
        ('Attributes', wintypes.WORD),
    )

class CONSOLE_SCREEN_BUFFER_INFO(Structure):
    _fields_ = (
        ('dwSize', wintypes._COORD),
        ('dwCursorPosition', wintypes._COORD),
        ('wAttributes', wintypes.WORD),
        ('srWindow', wintypes.SMALL_RECT),
        ('dwMaximumWindowSize', wintypes._COORD),
    )

class CONSOLE_CURSOR_INFO(Structure):
    _fields_ = (
        ("dwSize", wintypes.DWORD),
        ("bVisible", wintypes.BOOL),
    )

class CONSOLE_SCREEN_BUFFER_INFOEX(Structure):
    _fields_ = (
        ('cbSize', wintypes.ULONG),
        ('dwSize', wintypes._COORD),
        ('dwCursorPosition', wintypes._COORD),
        ('wAttributes', wintypes.WORD),
        ('srWindow', wintypes.SMALL_RECT),
        ('dwMaximumWindowSize', wintypes._COORD),
        ('wPopupAttributes', wintypes.WORD),
        ('bFullscreenSupported', wintypes.BOOL),
        ('ColorTable', wintypes.DWORD*16),
    )

class SECURITY_ATTRIBUTES(Structure):
    _fields_ = (
        ('nLength', wintypes.DWORD),
        ('lpSecurityDescriptor', wintypes.LPVOID),
        ('bInheritHandle', wintypes.BOOL),
    )

_GetStdHandle = windll.kernel32.GetStdHandle
_GetStdHandle.argtypes = (wintypes.DWORD,)
_GetStdHandle.restype = wintypes.HANDLE

_WriteConsoleW = windll.kernel32.WriteConsoleW
_WriteConsoleW.argtypes = (
    wintypes.HANDLE, wintypes.LPCWSTR, wintypes.DWORD, POINTER(wintypes.DWORD), wintypes.LPVOID
)

_WriteConsoleOutputW = windll.kernel32.WriteConsoleOutputW
_WriteConsoleOutputW.argtypes = (
    wintypes.HANDLE, POINTER(CHAR_INFO),
    wintypes._COORD, wintypes._COORD, POINTER(wintypes.SMALL_RECT)
)

_ReadConsoleInputW = windll.kernel32.ReadConsoleInputW
_ReadConsoleInputW.argtypes = (
    wintypes.HANDLE, POINTER(INPUT_RECORD), wintypes.DWORD, POINTER(wintypes.DWORD)
)

_GetNumberOfConsoleInputEvents = windll.kernel32.GetNumberOfConsoleInputEvents
_GetNumberOfConsoleInputEvents.argtypes = (wintypes.HANDLE, POINTER(wintypes.DWORD))

_GetConsoleScreenBufferInfo = windll.kernel32.GetConsoleScreenBufferInfo
_GetConsoleScreenBufferInfo.argtypes = (wintypes.HANDLE, POINTER(CONSOLE_SCREEN_BUFFER_INFO))

_SetConsoleTitleW = windll.kernel32.SetConsoleTitleW
_SetConsoleTitleW.argtypes = (wintypes.LPCWSTR,)
_SetConsoleTitleW.restype = wintypes.BOOL

_FillConsoleOutputCharacterW = windll.kernel32.FillConsoleOutputCharacterW
_FillConsoleOutputCharacterW.argtypes = (
    wintypes.HANDLE,
    wintypes.WCHAR,
    wintypes.DWORD,
    wintypes._COORD,
    POINTER(wintypes.DWORD),
)
_FillConsoleOutputCharacterW.restype = wintypes.BOOL

_FillConsoleOutputAttribute = windll.kernel32.FillConsoleOutputAttribute
_FillConsoleOutputAttribute.argtypes = (
    wintypes.HANDLE,
    wintypes.WORD,
    wintypes.DWORD,
    wintypes._COORD,
    POINTER(wintypes.DWORD),
)
_FillConsoleOutputAttribute.restype = wintypes.BOOL

_SetConsoleCursorPosition = windll.kernel32.SetConsoleCursorPosition
_SetConsoleCursorPosition.argtypes = (wintypes.HANDLE, wintypes._COORD)
_SetConsoleCursorPosition.restype = wintypes.BOOL

_GetConsoleCursorInfo = windll.kernel32.GetConsoleCursorInfo
_GetConsoleCursorInfo.argtypes = (wintypes.HANDLE, POINTER(CONSOLE_CURSOR_INFO))

_SetConsoleCursorInfo = windll.kernel32.SetConsoleCursorInfo
_SetConsoleCursorInfo.argtypes = (wintypes.HANDLE, POINTER(CONSOLE_CURSOR_INFO))

_SetConsoleScreenBufferInfoEx = windll.kernel32.SetConsoleScreenBufferInfoEx
_SetConsoleScreenBufferInfoEx.argtypes = (wintypes.HANDLE, POINTER(CONSOLE_SCREEN_BUFFER_INFOEX))

_GetConsoleScreenBufferInfoEx = windll.kernel32.GetConsoleScreenBufferInfoEx
_GetConsoleScreenBufferInfoEx.argtypes = (wintypes.HANDLE, POINTER(CONSOLE_SCREEN_BUFFER_INFOEX))

_ScrollConsoleScreenBuffer = windll.kernel32.ScrollConsoleScreenBufferW
_ScrollConsoleScreenBuffer.argtypes = (
    wintypes.HANDLE,
    POINTER(wintypes.SMALL_RECT),
    POINTER(wintypes.SMALL_RECT),
    wintypes._COORD,
    POINTER(CHAR_INFO),
)

_CreateConsoleScreenBuffer = windll.kernel32.CreateConsoleScreenBuffer
_CreateConsoleScreenBuffer.argtypes = (
    wintypes.DWORD,
    wintypes.DWORD,
    POINTER(SECURITY_ATTRIBUTES),
    wintypes.DWORD,
    wintypes.LPVOID
)
_CreateConsoleScreenBuffer.restype = wintypes.HANDLE

_SetConsoleActiveScreenBuffer = windll.kernel32.SetConsoleActiveScreenBuffer
_SetConsoleActiveScreenBuffer.argtypes = (wintypes.HANDLE,)
_SetConsoleActiveScreenBuffer.restype = wintypes.BOOL


_SetConsoleTextAttribute = windll.kernel32.SetConsoleTextAttribute
_SetConsoleTextAttribute.argtypes = (wintypes.HANDLE, wintypes.WORD)
_SetConsoleTextAttribute.restype = wintypes.BOOL

_SetConsoleScreenBufferSize = windll.kernel32.SetConsoleScreenBufferSize
_SetConsoleScreenBufferSize.argtypes = (wintypes.HANDLE, wintypes._COORD)

_SetConsoleWindowInfo = windll.kernel32.SetConsoleWindowInfo
_SetConsoleWindowInfo.argtypes = (
    wintypes.HANDLE,
    wintypes.BOOL,
    POINTER(wintypes.SMALL_RECT),
)

_SetConsoleMode = windll.kernel32.SetConsoleMode
_SetConsoleMode.argtypes = (
    wintypes.HANDLE,
    wintypes.DWORD,
)

_GetConsoleMode = windll.kernel32.GetConsoleMode
_GetConsoleMode.argtypes = (
    wintypes.HANDLE,
    POINTER(wintypes.DWORD),
)
_GetConsoleMode.restype = wintypes.BOOL


PHANDLER_ROUTINE = ctypes.WINFUNCTYPE(wintypes.BOOL, wintypes.DWORD)
_SetConsoleCtrlHandler = windll.kernel32.SetConsoleCtrlHandler
_SetConsoleCtrlHandler.argtypes = (
    PHANDLER_ROUTINE,
    wintypes.BOOL,
)
_SetConsoleCtrlHandler.restype = wintypes.BOOL


def GetConsoleScreenBufferInfo(handle):
    csbi = CONSOLE_SCREEN_BUFFER_INFO()
    _GetConsoleScreenBufferInfo(handle, byref(csbi))
    return csbi

def GetConsoleScreenBufferInfoEx(handle):
    csbie = CONSOLE_SCREEN_BUFFER_INFOEX()
    csbie.cbSize = wintypes.ULONG(ctypes.sizeof(csbie))
    _GetConsoleScreenBufferInfoEx(handle, byref(csbie))
    # work around Windows bug
    # see https://stackoverflow.com/questions/35901572/setconsolescreenbufferinfoex-bug
    csbie.srWindow.Bottom += 1
    csbie.srWindow.Right += 1
    return csbie

def FillConsoleOutputCharacter(handle, char, length, start):
    length = wintypes.DWORD(length)
    num_written = wintypes.DWORD(0)
    _FillConsoleOutputCharacterW(handle, char, length, start, byref(num_written))
    return num_written.value

def FillConsoleOutputAttribute(handle, attr, length, start):
    attribute = wintypes.WORD(attr)
    length = wintypes.DWORD(length)
    return _FillConsoleOutputAttribute(handle, attribute, length, start, byref(wintypes.DWORD()))

def GetConsoleCursorInfo(handle):
    cci = CONSOLE_CURSOR_INFO()
    _GetConsoleCursorInfo(handle, byref(cci))
    return cci

def SetConsoleCursorInfo(handle, cci):
    _SetConsoleCursorInfo(handle, byref(cci))

def ScrollConsoleScreenBuffer(handle, scroll_rect, clip_rect, new_position, char, attr):
    char_info = CHAR_INFO(char, attr)
    _ScrollConsoleScreenBuffer(
        handle, byref(scroll_rect), byref(clip_rect), new_position, byref(char_info)
    )

def GetConsoleMode(handle):
    mode = wintypes.DWORD()
    _GetConsoleMode(handle, mode)
    return mode


# standard stream constants
# these remain unchanged if we switch to another screen buffer
HSTDIN = _GetStdHandle(-10)
HSTDOUT = _GetStdHandle(-11)
HSTDERR = _GetStdHandle(-12)


class _ConsoleWriter:
    """Singleton to manage writing to console and consequent scrolling."""

    _overflow = False

    @classmethod
    def write(cls, handle, unistr):
        """Write character to console, avoid scroll on bottom line."""
        csbi = CONSOLE_SCREEN_BUFFER_INFO()
        _GetConsoleScreenBufferInfo(handle, byref(csbi))
        col, row = csbi.dwCursorPosition.X, csbi.dwCursorPosition.Y
        width, height = csbi.dwSize.X, csbi.dwSize.Y
        for ch in unistr:
            if (col >= width - 1 and ch not in (u'\n', u'\b', u'\r') and not cls._overflow):
                ci = CHAR_INFO(ch, csbi.wAttributes)
                # do not advance cursor if we're on the last position of the
                # screen buffer, to avoid unwanted scrolling.
                _WriteConsoleOutputW(
                    handle, byref(ci), wintypes._COORD(1, 1), wintypes._COORD(0, 0),
                    wintypes.SMALL_RECT(col, row, col, row)
                )
            else:
                if cls._overflow and ch not in (u'\n', u'\r', u'\b'):
                    _WriteConsoleW(
                        handle, u'\r\n', 1,
                        byref(wintypes.DWORD()), byref(wintypes.DWORD())
                    )
                    col = 0
                    cls._overflow = False
                _WriteConsoleW(
                    handle, ch, 1,
                    byref(wintypes.DWORD()), byref(wintypes.DWORD())
                )
            # if we don't include \n here we get glitches on regular console writes
            # is it necessary to treat CR and LF separately *in raw mode*?
            # i.e. in raw mode,  shouldn't LF just move a line down without changing the column?
            if ch in (u'\r', u'\n'):
                col = 0
                cls._overflow = False
            elif ch == u'\b':
                col = max(col-1, 0)
                cls._overflow = False
            else:
                col = min(col+1, width-1)


##############################################################################
# console class

@PHANDLER_ROUTINE
def _ctrl_handler(fdwCtrlType):
    """Handle Ctrl-Break event."""
    # CTRL_BREAK_EVENT
    return (fdwCtrlType == 1)


class Win32Console(object):
    """Win32API-based console implementation."""

    def __init__(self):
        """Set up console"""
        # preserve original settings
        self._orig_csbie = GetConsoleScreenBufferInfoEx(HSTDOUT)
        self._orig_stdin_mode = GetConsoleMode(HSTDIN)
        self._attrs = self._orig_csbie.wAttributes
        # input
        self._input_buffer = deque()
        self._echo = True
        self._save_stdout = None
        # standard streams - these may change in alternative screen buffe rmode
        self._hstdin = HSTDIN
        self._hstdout = HSTDOUT
        self._hstderr = HSTDERR

    def set_raw(self):
        """Enter raw terminal mode (no echo, don't exit on ctrl-C)."""
        self._echo = False
        # unset ENABLE_PROCESSED_INPUT
        _SetConsoleMode(self._hstdin, wintypes.DWORD(self._orig_stdin_mode.value & ~0x0001))
        # don't exit on ctrl-Break
        _SetConsoleCtrlHandler(_ctrl_handler, True)

    def unset_raw(self):
        """Leave raw terminal mode."""
        self._echo = True
        _SetConsoleMode(self._hstdin, self._orig_stdin_mode)

    def start_screen(self):
        """Enter full-screen/application mode."""
        # https://docs.microsoft.com/en-us/windows/console/reading-and-writing-blocks-of-characters-and-attributes
        new_buffer = _CreateConsoleScreenBuffer(
            wintypes.DWORD(0xc0000000), # GENERIC_READ | GENERIC_WRITE
            wintypes.DWORD(0x3), # FILE_SHARE_READ | FILE_SHARE_WRITE
            None,
            wintypes.DWORD(1), # CONSOLE_TEXTMODE_BUFFER
            None
        )
        _SetConsoleActiveScreenBuffer(new_buffer)
        self._save_stdout, self._hstdout = self._hstdout, new_buffer
        self.set_raw()

    def close_screen(self):
        """Leave full-screen/application mode."""
        self.unset_raw()
        self.reset()
        self._hstdout = self._save_stdout
        _SetConsoleActiveScreenBuffer(self._hstdout)

    def key_pressed(self):
        """key pressed on keyboard."""
        return msvcrt.kbhit()

    def set_caption(self, caption):
        """Set terminal caption."""
        _SetConsoleTitleW(caption)

    def resize(self, height, width):
        """Resize terminal."""
        csbi = GetConsoleScreenBufferInfo(self._hstdout)
        # SetConsoleScreenBufferSize can't make the buffer smaller than the window
        # SetConsoleWindowInfo can't make the window larger than the buffer (in either direction)
        # allow for both shrinking and growing by calling one of them twice,
        # for each direction separately
        new_size = wintypes._COORD(width, csbi.dwSize.Y)
        new_window = wintypes.SMALL_RECT(0, 0, width-1, csbi.dwSize.Y-1)
        _SetConsoleScreenBufferSize(self._hstdout, new_size)
        _SetConsoleWindowInfo(self._hstdout, True, new_window)
        _SetConsoleScreenBufferSize(self._hstdout, new_size)
        new_size = wintypes._COORD(width, height)
        new_window = wintypes.SMALL_RECT(0, 0, width-1, height-1)
        _SetConsoleScreenBufferSize(self._hstdout, new_size)
        _SetConsoleWindowInfo(self._hstdout, True, new_window)
        _SetConsoleScreenBufferSize(self._hstdout, new_size)

    ##########################################################################
    # output

    def write(self, unistr):
        """Write (unicode) text to the console."""
        _ConsoleWriter.write(self._hstdout, unistr)

    def clear(self):
        """Clear the screen."""
        csbi = GetConsoleScreenBufferInfo(self._hstdout)
        # fill the entire screen with blanks
        FillConsoleOutputCharacter(
            self._hstdout, u' ', csbi.dwSize.X * csbi.dwSize.Y, wintypes._COORD(0, 0)
        )
        # now set the buffer's attributes accordingly
        FillConsoleOutputAttribute(
            self._hstdout, self._attrs, csbi.dwSize.X * csbi.dwSize.Y, wintypes._COORD(0, 0)
        )
        _SetConsoleCursorPosition(self._hstdout, wintypes._COORD(0, 0))

    def clear_row(self, width=None):
        """Clear the current row."""
        csbi = GetConsoleScreenBufferInfo(self._hstdout)
        from_coord = wintypes._COORD(0, csbi.dwCursorPosition.Y)
        # fill the entire row with blanks
        if width is None:
            width = csbi.dwSize.X
        FillConsoleOutputCharacter(self._hstdout, u' ', width, from_coord)
        # now set the buffer's attributes accordingly
        FillConsoleOutputAttribute(self._hstdout, self._attrs, width, from_coord)

    def set_cursor_colour(self, colour):
        """Set the current cursor colour attribute - not supported."""

    def show_cursor(self, block=False):
        """Show the cursor."""
        curs_info = GetConsoleCursorInfo(self._hstdout)
        curs_info.bVisible = True
        curs_info.dwSize = 100 if block else 20
        SetConsoleCursorInfo(self._hstdout, curs_info)

    def hide_cursor(self):
        """Hide the cursor."""
        curs_info = GetConsoleCursorInfo(self._hstdout)
        curs_info.bVisible = False
        SetConsoleCursorInfo(self._hstdout, curs_info)

    def move_cursor_to(self, row, col):
        """Move cursor to a new position (1,1 is top left)."""
        csbi = GetConsoleScreenBufferInfo(self._hstdout)
        row, col = row-1, col-1
        while col >= csbi.dwSize.X:
            col -= csbi.dwSize.X
            row += 1
        while col < 0:
            col += csbi.dwSize.X
            row -= 1
        # If the position is out of range, do nothing.
        if row >= 0 and col >= 0:
            _SetConsoleCursorPosition(self._hstdout, wintypes._COORD(col, row))

    def scroll(self, top, bottom, rows):
        """Scroll the region between top and bottom one row up (-) or down (+)."""
        if not rows:
            return
        # use zero-based indexing
        start, stop = top-1, bottom-1
        # we're using opposuite sign conventions
        csbi = GetConsoleScreenBufferInfo(self._hstdout)
        # absolute position of window in screen buffer
        # interpret other coordinates as relative to the window
        window = csbi.srWindow
        # scroll region
        clip_rect = wintypes.SMALL_RECT(
            window.Left, window.Top + start, window.Right, window.Top + stop
        )
        if rows < 0:
            # minus signs since rows is a negative number
            region = wintypes.SMALL_RECT(window.Left, window.Top - rows, window.Right, window.Bottom)
            new_pos = wintypes._COORD(window.Left, window.Top)
        else:
            region = wintypes.SMALL_RECT(window.Left, window.Top, window.Right, window.Bottom + rows)
            new_pos = wintypes._COORD(window.Left, window.Top + rows)
        # workaround: in this particular case, Windows doesn't seem to respect the clip area.
        if (
                clip_rect.Bottom == window.Bottom-1 and
                region.Bottom >= window.Bottom-1 and
                new_pos.Y < region.Top
            ):
            # first scroll everything up
            clip_rect.Bottom = window.Bottom
            bottom, region.Bottom = region.Bottom, window.Bottom
            ScrollConsoleScreenBuffer(self._hstdout, region, clip_rect, new_pos, u' ', self._attrs)
            # and then scroll the bottom back down
            new_pos.Y = window.Bottom
            region.Top = bottom-1
            ScrollConsoleScreenBuffer(self._hstdout, region, clip_rect, new_pos, u' ', self._attrs)
        else:
            ScrollConsoleScreenBuffer(self._hstdout, region, clip_rect, new_pos, u' ', self._attrs)

    def reset(self):
        """Reset to default attributes."""
        _SetConsoleScreenBufferInfoEx(self._hstdout, byref(self._orig_csbie))
        self._attrs = self._orig_csbie.wAttributes
        self.show_cursor()

    def set_attributes(self, fore, back, blink, underline):
        """Set current attributes."""
        self._attrs = fore % 8 + back * 16 + (BRIGHT if fore > 8 else NORMAL)
        _SetConsoleTextAttribute(self._hstdout, self._attrs)

    def set_palette_entry(self, attr, red, green, blue):
        """Set palette entry for attribute (0--16)."""
        csbie = GetConsoleScreenBufferInfoEx(self._hstdout)
        csbie.ColorTable[attr] = (
            0x00010000 * blue + 0x00000100 * green + 0x00000001 * red
        )
        _SetConsoleScreenBufferInfoEx(self._hstdout, byref(csbie))

    ##########################################################################
    # input

    def read_key(self):
        """
        Read keypress from console. Non-blocking.
        Returns tuple (unicode, keycode, set of mods)
        """
        self._fill_buffer(blocking=False)
        if not self._input_buffer:
            return (u'', None, {})
        return self._input_buffer.popleft()

    def read_all_chars(self):
        """Read all characters in the buffer."""
        self._fill_buffer(blocking=False)
        closed = False
        if self._input_buffer and self._input_buffer[-1][0] == u'\x1a':
            closed = True
            self._input_buffer.pop()
            if not self._input_buffer:
                return None
        output = u''.join(
            _char for _char, _, _ in self._input_buffer
        )
        self._input_buffer.clear()
        if closed:
            self._input_buffer.append(u'\x1a')
        return output

    def _fill_buffer(self, blocking):
        """Interpret all key events."""
        while True:
            nevents = wintypes.DWORD()
            _GetNumberOfConsoleInputEvents(self._hstdin, byref(nevents))
            if not nevents.value and not blocking:
                return
            # only ever block on first loop
            blocking = False
            if nevents.value > 0:
                input_buffer = (INPUT_RECORD * nevents.value)()
                nread = wintypes.DWORD()
                _ReadConsoleInputW(
                    self._hstdin,
                    cast(input_buffer, POINTER(INPUT_RECORD)),
                    nevents.value, byref(nread)
                )
                for event in input_buffer:
                    if event.EventType != 1: # KEY_EVENT
                        continue
                    char, key, mods = self._translate_event(event)
                    if char or key:
                        self._input_buffer.append((char, key, mods))
                        if char == u'\x1a':
                            # ctrl-z is end of input on windows console
                            return
                        if self._echo:
                            _ConsoleWriter.write(self._hstdout, char.replace(u'\r', u'\n'))
            time.sleep(0.01)

    def _translate_event(self, event):
        char = event.KeyEvent.UnicodeChar
        key = event.KeyEvent.wVirtualKeyCode
        control = event.KeyEvent.dwControlKeyState
        if char == u'\0':
            # windows uses null-terminated strings so \0 means no output
            char = u''
        if not event.KeyEvent.bKeyDown:
            # key-up event for unicode Alt+HEX input
            if event.KeyEvent.wVirtualKeyCode == KEYS.ALT:
                return char, None, set()
            # ignore other key-up events
            return u'', None, set()
        # decode modifier bit flags
        mods = set(key for key, mask in MODS.items() if control & mask)
        key = VK_TO_KEY.get(key, None)
        return char, key, mods


def _has_console():
    """Determine if we have a console attached or are a GUI app."""
    try:
        return bool(_GetConsoleMode(HSTDOUT, byref(wintypes.DWORD())))
    except Exception as e:
        return False


IS_CONSOLE_APP = _has_console()


##############################################################################
# standard i/o

if PY2: # pragma: no cover

    class _ConsoleOutput(StreamWrapper):
        """Bytes stream wrapper using Unicode API, to replace Python2 sys.stdout."""

        def __init__(self, stream, handle, encoding='utf-8'):
            StreamWrapper.__init__(self, stream)
            self._handle = handle
            self.encoding = encoding

        def write(self, bytestr):
            if not isinstance(bytestr, bytes):
                raise TypeError('write() argument must be bytes, not %s' % type(bytestr))
            unistr = bytestr.decode(self.encoding, errors='replace')
            _ConsoleWriter.write(self._handle, unistr)


    class _ConsoleInput(StreamWrapper):
        """Bytes stream wrapper using Unicode API, to replace Python2 sys.stdin."""

        def __init__(self, console, encoding='utf-8'):
            StreamWrapper.__init__(self, sys.stdin)
            self._handle = HSTDIN
            self._console = console
            self.encoding = encoding

        def read(self, size=-1):
            output = bytearray()
            while size < 0 or len(output) < size:
                key = self._console.read_key()
                if isinstance(key, int):
                    continue
                output.append(key.encode(self.encoding, errors='replace'))
            return bytes(output)


class StdIO(StdIOBase):
    """Holds standard unicode streams."""

    def __init__(self, console):
        self._console = console
        StdIOBase.__init__(self)

    if PY2: # pragma: no cover

        def _attach_stdin(self):
            if sys.stdin.isatty():
                self.stdin = self._wrap_input_stream(_ConsoleInput(self._console))
            else:
                self.stdin = self._wrap_input_stream(sys.stdin)

        def _attach_output_stream(self, stream_name, redirected=False):
            stream = getattr(sys, '__%s__' % (stream_name,))
            if stream.isatty() and not redirected:
                handle = {'stdout': HSTDOUT, 'stderr': HSTDERR}[stream_name]
                new_stream = self._wrap_output_stream(_ConsoleOutput(stream, handle))
            else:
                encoding = 'utf-8' if redirected else None
                new_stream = self._wrap_output_stream(stream, encoding)
            setattr(self, stream_name, new_stream)


def init_stdio():
    """Platform-specific initialisation."""
    # set stdio as binary, to avoid Windows messing around with CRLFs
    # only do this for redirected output, as it breaks interactive Python sessions
    try:
        if not sys.stdin.isatty():
            msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
        if not sys.stdout.isatty():
            msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)
    except EnvironmentError:
        # raises an error if started in gui mode, as we have no stdio
        pass



"""
PC-BASIC - statements.py
Statement parser

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging
import struct
from functools import partial

from ...compat import iterchar
from ..base import error
from ..base import tokens as tk
from ..base.tokens import DIGITS, LETTERS
from .. import values
from . import expressions
from . import userfunctions


class Parser(object):
    """BASIC statement parser."""

    def __init__(self, values, memory, syntax):
        """Initialise statement context."""
        # re-execute current statement after Break
        self.redo_on_break = False
        # expression parser
        self.expression_parser = expressions.ExpressionParser(values, memory)
        self.user_functions = self.expression_parser.user_functions
        # syntax: advanced, pcjr, tandy
        self._syntax = syntax
        # initialise syntax parser tables
        self._init_syntax()

    def __getstate__(self):
        """Pickle."""
        pickle_dict = self.__dict__.copy()
        # can't be pickled
        pickle_dict['_simple'] = None
        pickle_dict['_complex'] = None
        pickle_dict['_callbacks'] = None
        return pickle_dict

    def __setstate__(self, pickle_dict):
        """Unpickle."""
        self.__dict__.update(pickle_dict)
        self._init_syntax()

    def init_callbacks(self, session):
        """Assign statement and function callbacks."""
        self.init_statements(session)
        self.expression_parser.init_functions(session)

    def parse_statement(self, ins):
        """Parse and execute a single statement."""
        # read keyword token or one byte
        ins.skip_blank()
        c = ins.read_keyword_token()
        if c in self._simple:
            parse_args = self._simple[c]
        elif c in self._complex:
            stat_dict = self._complex[c]
            ins.skip_blank()
            selector = ins.read_keyword_token()
            ins.seek(-len(selector), 1)
            if selector not in stat_dict.keys():
                selector = None
            else:
                c += selector
            parse_args = stat_dict[selector]
        else:
            ins.seek(-len(c), 1)
            if c in set(iterchar(LETTERS)):
                # implicit LET
                c = tk.LET
                parse_args = self._simple[tk.LET]
            else:
                ins.require_end()
                return
        self._callbacks[c](parse_args(ins))
        # end-of-statement is checked at start of next statement in interpreter loop

    def parse_name(self, ins):
        """Get scalar part of variable name from token stream."""
        name = ins.read_name()
        # must not be empty
        error.throw_if(not name, error.STX)
        # append sigil, if missing
        return name

    def parse_expression(self, ins, allow_empty=False):
        """Compute the value of the expression at the current code pointer."""
        if allow_empty and ins.skip_blank() in tk.END_EXPRESSION:
            return None
        self.redo_on_break = True
        val = self.expression_parser.parse_expression(ins)
        self.redo_on_break = False
        return val

    ###########################################################################

    def _init_syntax(self):
        """Initialise syntax parsers."""
        self._simple = {
            tk.DATA: self._skip_statement,
            tk.COMMON: self._skip_statement,
            tk.REM: self._skip_line,
            tk.ELSE: self._skip_line,
            tk.CONT: self._parse_nothing,
            tk.TRON: self._parse_nothing,
            tk.TROFF: self._parse_nothing,
            tk.WHILE: self._parse_nothing,
            tk.RESET: self._parse_end,
            tk.END: self._parse_end,
            tk.STOP: self._parse_end,
            tk.NEW: self._parse_end,
            tk.WEND: self._parse_end,
            tk.SYSTEM: self._parse_end,
            tk.FOR: self._parse_for,
            tk.NEXT: self._parse_next,
            tk.INPUT: self._parse_input,
            tk.DIM: self._parse_var_list,
            tk.READ: self._parse_var_list,
            tk.LET: self._parse_let,
            tk.GOTO: self._parse_single_line_number,
            tk.RUN: self._parse_run,
            tk.IF: self._parse_if,
            tk.RESTORE: self._parse_restore,
            tk.GOSUB: self._parse_single_line_number,
            tk.RETURN: self._parse_optional_line_number,
            tk.PRINT: partial(self._parse_print, parse_file=True),
            tk.CLEAR: self._parse_clear,
            tk.LIST: self._parse_list,
            tk.WAIT: self._parse_wait,
            tk.POKE: self._parse_two_args,
            tk.OUT: self._parse_two_args,
            tk.LPRINT: partial(self._parse_print, parse_file=False),
            tk.LLIST: self._parse_delete_llist,
            tk.WIDTH: self._parse_width,
            tk.SWAP: self._parse_swap,
            tk.ERASE: self._parse_erase,
            tk.EDIT: self._parse_edit,
            tk.ERROR: self._parse_single_arg,
            tk.RESUME: self._parse_resume,
            tk.DELETE: self._parse_delete_llist,
            tk.AUTO: self._parse_auto,
            tk.RENUM: self._parse_renum,
            tk.DEFSTR: self._parse_deftype,
            tk.DEFINT: self._parse_deftype,
            tk.DEFSNG: self._parse_deftype,
            tk.DEFDBL: self._parse_deftype,
            tk.CALL: self._parse_call,
            tk.CALLS: self._parse_call,
            tk.WRITE: self._parse_write,
            tk.OPTION: self._parse_option_base,
            tk.RANDOMIZE: self._parse_optional_arg,
            tk.OPEN: self._parse_open,
            tk.CLOSE: self._parse_close,
            tk.LOAD: self._parse_load,
            tk.MERGE: self._parse_single_arg_no_end,
            tk.SAVE: self._parse_save,
            tk.COLOR: self._parse_color,
            tk.CLS: self._parse_cls,
            tk.MOTOR: self._parse_optional_arg,
            tk.BSAVE: self._parse_bsave,
            tk.BLOAD: self._parse_bload,
            tk.SOUND: self._parse_sound,
            tk.BEEP: self._parse_beep,
            tk.PSET: self._parse_pset_preset,
            tk.PRESET: self._parse_pset_preset,
            tk.SCREEN: self._parse_screen,
            tk.LOCATE: self._parse_locate,
            tk.FILES: self._parse_optional_arg_no_end,
            tk.FIELD: self._parse_field,
            tk.NAME: self._parse_name,
            tk.LSET: self._parse_let,
            tk.RSET: self._parse_let,
            tk.KILL: self._parse_single_arg_no_end,
            tk.CHAIN: self._parse_chain,
            tk.DATE: self._parse_time_date,
            tk.TIME: self._parse_time_date,
            tk.PAINT: self._parse_paint,
            tk.COM: self._parse_com_command,
            tk.CIRCLE: self._parse_circle,
            tk.DRAW: self._parse_single_arg,
            tk.TIMER: self._parse_event_command,
            tk.IOCTL: self._parse_ioctl,
            tk.CHDIR: self._parse_single_arg_no_end,
            tk.MKDIR: self._parse_single_arg_no_end,
            tk.RMDIR: self._parse_single_arg_no_end,
            tk.SHELL: self._parse_optional_arg_no_end,
            tk.ENVIRON: self._parse_single_arg_no_end,
            tk.WINDOW: self._parse_window,
            tk.LCOPY: self._parse_optional_arg,
            tk.PCOPY: self._parse_pcopy,
            tk.LOCK: self._parse_lock_unlock,
            tk.UNLOCK: self._parse_lock_unlock,
            tk.MID: self._parse_mid,
            tk.PEN: self._parse_event_command,
            b'_': self._parse_call_extension,
        }
        if self._syntax in ('pcjr', 'tandy'):
            self._simple.update({
                tk.TERM: self._parse_end,
                tk.NOISE: self._parse_noise,
            })
        self._complex = {
            tk.ON: {
                tk.ERROR: self._parse_on_error_goto,
                tk.KEY: self._parse_on_event,
                tk.PEN: self._parse_on_event,
                tk.TIMER: self._parse_on_event,
                tk.PLAY: self._parse_on_event,
                tk.COM: self._parse_on_event,
                tk.STRIG: self._parse_on_event,
                None: self._parse_on_jump,
            },
            tk.DEF: {
                tk.FN: self._parse_def_fn,
                tk.USR: self._parse_def_usr,
                None: self._parse_def_seg,
            },
            tk.LINE: {
                tk.INPUT: self._parse_line_input,
                None: self._parse_line,
            },
            tk.KEY: {
                tk.ON: self._parse_key_macro,
                tk.OFF: self._parse_key_macro,
                tk.LIST: self._parse_key_macro,
                b'(': self._parse_com_command,
                None: self._parse_two_args,
            },
            tk.PUT: {
                b'(': self._parse_put_graph,
                None: self._parse_put_get_file,
            },
            tk.GET: {
                b'(': self._parse_get_graph,
                None: self._parse_put_get_file,
            },
            tk.PLAY: {
                tk.ON: self._parse_event_command,
                tk.OFF: self._parse_event_command,
                tk.STOP: self._parse_event_command,
                None: self._parse_play,
            },
            tk.VIEW: {
                tk.PRINT: self._parse_view_print,
                None: self._parse_view,
            },
            tk.PALETTE: {
                tk.USING: self._parse_palette_using,
                None: self._parse_palette,
            },
            tk.STRIG: {
                tk.ON: self._parse_strig_switch,
                tk.OFF: self._parse_strig_switch,
                None: self._parse_com_command,
            },
        }

    def init_statements(self, session):
        """Initialise statement callbacks."""
        self._callbacks = {
            tk.DATA: list,
            tk.COMMON: list,
            tk.REM: list,
            tk.ELSE: list,
            tk.CONT: session.interpreter.cont_,
            tk.TRON: session.interpreter.tron_,
            tk.TROFF: session.interpreter.troff_,
            tk.WHILE: session.interpreter.while_,
            tk.RESET: session.files.reset_,
            tk.END: session.end_,
            tk.STOP: session.interpreter.stop_,
            tk.NEW: session.new_,
            tk.WEND: session.interpreter.wend_,
            tk.SYSTEM: session.system_,
            tk.FOR: session.interpreter.for_,
            tk.NEXT: session.interpreter.next_,
            tk.INPUT: session.input_,
            tk.DIM: session.memory.arrays.dim_,
            tk.READ: session.interpreter.read_,
            tk.LET: session.memory.let_,
            tk.GOTO: session.interpreter.goto_,
            tk.RUN: session.run_,
            tk.IF: session.interpreter.if_,
            tk.RESTORE: session.interpreter.restore_,
            tk.GOSUB: session.interpreter.gosub_,
            tk.RETURN: session.interpreter.return_,
            tk.PRINT: session.files.print_,
            tk.CLEAR: session.clear_,
            tk.LIST: session.list_,
            tk.WAIT: session.machine.wait_,
            tk.POKE: session.all_memory.poke_,
            tk.OUT: session.machine.out_,
            tk.LPRINT: session.files.lprint_,
            tk.LLIST: session.interpreter.llist_,
            tk.WIDTH: session.files.width_,
            tk.SWAP: session.memory.swap_,
            tk.ERASE: session.memory.arrays.erase_,
            tk.EDIT: session.edit_,
            tk.ERROR: session.interpreter.error_,
            tk.RESUME: session.interpreter.resume_,
            tk.DELETE: session.delete_,
            tk.AUTO: session.auto_,
            tk.RENUM: session.interpreter.renum_,
            tk.DEFSTR: session.memory.defstr_,
            tk.DEFINT: session.memory.defint_,
            tk.DEFSNG: session.memory.defsng_,
            tk.DEFDBL: session.memory.defdbl_,
            tk.CALL: session.all_memory.call_,
            tk.CALLS: session.all_memory.call_,
            tk.WRITE: session.files.write_,
            tk.OPTION: session.memory.arrays.option_base_,
            tk.RANDOMIZE: session.randomize_,
            tk.OPEN: session.files.open_,
            tk.CLOSE: session.files.close_,
            tk.LOAD: session.load_,
            tk.MERGE: session.merge_,
            tk.SAVE: session.save_,
            tk.COLOR: session.display.color_,
            tk.CLS: session.display.cls_,
            tk.MOTOR: session.files.motor_,
            tk.BSAVE: session.all_memory.bsave_,
            tk.BLOAD: session.all_memory.bload_,
            tk.SOUND: session.sound.sound_,
            tk.BEEP: session.sound.beep_,
            tk.PSET: session.graphics.pset_,
            tk.PRESET: session.graphics.preset_,
            tk.SCREEN: session.display.screen_,
            tk.LOCATE: session.text_screen.locate_,
            tk.FILES: session.files.files_,
            tk.FIELD: session.files.field_,
            tk.NAME: session.files.name_,
            tk.LSET: session.memory.lset_,
            tk.RSET: session.memory.rset_,
            tk.KILL: session.files.kill_,
            tk.CHAIN: session.chain_,
            tk.DATE: session.clock.date_,
            tk.TIME: session.clock.time_,
            tk.PAINT: session.graphics.paint_,
            tk.COM: session.basic_events.com_,
            tk.CIRCLE: session.graphics.circle_,
            tk.DRAW: session.graphics.draw_,
            tk.TIMER: session.basic_events.timer_,
            tk.IOCTL: session.files.ioctl_statement_,
            tk.CHDIR: session.files.chdir_,
            tk.MKDIR: session.files.mkdir_,
            tk.RMDIR: session.files.rmdir_,
            tk.SHELL: session.shell_,
            tk.ENVIRON: session.environment.environ_statement_,
            tk.WINDOW: session.graphics.window_,
            tk.LCOPY: session.files.lcopy_,
            tk.PCOPY: session.display.pcopy_,
            tk.LOCK: session.files.lock_,
            tk.UNLOCK: session.files.unlock_,
            tk.MID: session.memory.mid_,
            tk.PEN: session.basic_events.pen_,
            tk.TERM: session.term_,
            tk.NOISE: session.sound.noise_,
            tk.ON + tk.ERROR: session.interpreter.on_error_goto_,
            tk.ON + tk.KEY: session.basic_events.on_event_gosub_,
            tk.ON + tk.PEN: session.basic_events.on_event_gosub_,
            tk.ON + tk.TIMER: session.basic_events.on_event_gosub_,
            tk.ON + tk.PLAY: session.basic_events.on_event_gosub_,
            tk.ON + tk.COM: session.basic_events.on_event_gosub_,
            tk.ON + tk.STRIG: session.basic_events.on_event_gosub_,
            tk.ON: session.interpreter.on_jump_,
            tk.DEF + tk.FN: session.interpreter.def_fn_,
            tk.DEF + tk.USR: session.all_memory.def_usr_,
            tk.DEF: session.all_memory.def_seg_,
            tk.LINE + tk.INPUT: session.line_input_,
            tk.LINE: session.graphics.line_,
            tk.KEY + tk.ON: session.console.key_,
            tk.KEY + tk.OFF: session.console.key_,
            tk.KEY + tk.LIST: session.console.key_,
            tk.KEY + b'(': session.basic_events.key_,
            tk.KEY: session.key_,
            tk.PUT + b'(': session.graphics.put_,
            tk.PUT: session.files.put_,
            tk.GET + b'(': session.graphics.get_,
            tk.GET: session.files.get_,
            tk.PLAY + tk.ON: session.basic_events.play_,
            tk.PLAY + tk.OFF: session.basic_events.play_,
            tk.PLAY + tk.STOP: session.basic_events.play_,
            tk.PLAY: session.sound.play_,
            tk.VIEW + tk.PRINT: session.text_screen.view_print_,
            tk.VIEW: session.graphics.view_,
            tk.PALETTE + tk.USING: session.display.palette_using_,
            tk.PALETTE: session.display.palette_,
            tk.STRIG + tk.ON: session.stick.strig_statement_,
            tk.STRIG + tk.OFF: session.stick.strig_statement_,
            tk.STRIG: session.basic_events.strig_,
            b'_': session.extensions.call_as_statement,
        }

    ###########################################################################
    # auxiliary functions

    def _parse_bracket(self, ins):
        """Compute the value of the bracketed expression."""
        ins.require_read((b'(',))
        # we'll get a Syntax error, not a Missing operand, if we close with )
        val = self.parse_expression(ins)
        ins.require_read((b')',))
        return val

    def _parse_variable(self, ins):
        """Helper function: parse a scalar or array element."""
        name = ins.read_name()
        error.throw_if(not name, error.STX)
        self.redo_on_break = True
        indices = self.expression_parser.parse_indices(ins)
        self.redo_on_break = False
        return name, indices

    def _parse_jumpnum(self, ins):
        """Parses a line number pointer as in GOTO, GOSUB, LIST, RENUM, EDIT, etc."""
        ins.require_read((tk.T_UINT,))
        token = ins.read(2)
        assert len(token) == 2, 'Bytecode truncated in line number pointer'
        return struct.unpack('<H', token)[0]

    def _parse_optional_jumpnum(self, ins):
        """Parses a line number pointer as in GOTO, GOSUB, LIST, RENUM, EDIT, etc."""
        # no line number
        if ins.skip_blank() != tk.T_UINT:
            return None
        return self._parse_jumpnum(ins)

    def _parse_line_range(self, ins):
        """Parse a line number range as in LIST, DELETE."""
        from_line = self._parse_jumpnum_or_dot(ins, allow_empty=True)
        if ins.skip_blank_read_if((tk.O_MINUS,)):
            to_line = self._parse_jumpnum_or_dot(ins, allow_empty=True)
        else:
            to_line = from_line
        return (from_line, to_line)

    def _parse_jumpnum_or_dot(self, ins, allow_empty=False, err=error.STX):
        """Parse jump target; returns int, None or '.'"""
        c = ins.skip_blank_read()
        if c == tk.T_UINT:
            token = ins.read(2)
            assert len(token) == 2, 'bytecode truncated in line number pointer'
            return struct.unpack('<H', token)[0]
        elif c == b'.':
            return b'.'
        else:
            if allow_empty:
                ins.seek(-len(c), 1)
                return None
            raise error.BASICError(err)

    ###########################################################################
    # no arguments

    def _parse_nothing(self, ins):
        """Parse nothing."""
        # e.g. TRON LAH raises error but TRON will have been executed
        return
        yield # pragma: no cover

    def _parse_end(self, ins):
        """Parse end-of-statement before executing argumentless statement."""
        # e.g. SYSTEM LAH does not execute
        ins.require_end()
        # empty generator
        return
        yield # pragma: no cover

    def _skip_line(self, ins):
        """Ignore the rest of the line."""
        ins.skip_to(tk.END_LINE)
        return
        yield # pragma: no cover

    def _skip_statement(self, ins):
        """Ignore rest of statement."""
        ins.skip_to(tk.END_STATEMENT)
        return
        yield # pragma: no cover

    ###########################################################################
    # single argument

    def _parse_optional_arg(self, ins):
        """Parse statement with one optional argument."""
        yield self.parse_expression(ins, allow_empty=True)
        ins.require_end()

    def _parse_optional_arg_no_end(self, ins):
        """Parse statement with one optional argument."""
        yield self.parse_expression(ins, allow_empty=True)

    def _parse_single_arg(self, ins):
        """Parse statement with one mandatory argument."""
        yield self.parse_expression(ins)
        ins.require_end()

    def _parse_single_arg_no_end(self, ins):
        """Parse statement with one mandatory argument."""
        yield self.parse_expression(ins)

    def _parse_single_line_number(self, ins):
        """Parse statement with single line number argument."""
        yield self._parse_jumpnum(ins)

    def _parse_optional_line_number(self, ins):
        """Parse statement with optional line number argument."""
        jumpnum = None
        if ins.skip_blank() == tk.T_UINT:
            jumpnum = self._parse_jumpnum(ins)
        yield jumpnum

    ###########################################################################
    # two arguments

    def _parse_two_args(self, ins):
        """Parse POKE or OUT syntax."""
        yield self.parse_expression(ins)
        ins.require_read((b',',))
        yield self.parse_expression(ins)

    ###########################################################################
    # flow-control statements

    def _parse_run(self, ins):
        """Parse RUN syntax."""
        c = ins.skip_blank()
        if c == tk.T_UINT:
            # parse line number and ignore rest of line
            yield self._parse_jumpnum(ins)
        elif c not in tk.END_STATEMENT:
            yield None
            yield self.parse_expression(ins)
            if ins.skip_blank_read_if((b',',)):
                ins.require_read((b'R',))
                yield True
            else:
                yield False
        else:
            yield None

    def _parse_resume(self, ins):
        """Parse RESUME syntax."""
        c = ins.skip_blank()
        if c == tk.NEXT:
            yield ins.read(1)
        elif c in tk.END_STATEMENT:
            yield None
        else:
            yield self._parse_jumpnum(ins)
        ins.require_end()

    def _parse_on_error_goto(self, ins):
        """Parse ON ERROR GOTO syntax."""
        ins.require_read((tk.ERROR,))
        ins.require_read((tk.GOTO,))
        yield self._parse_jumpnum(ins)

    ###########################################################################
    # event statements

    def _parse_event_command(self, ins):
        """Parse PEN, PLAY or TIMER syntax."""
        yield ins.require_read((tk.ON, tk.OFF, tk.STOP))

    def _parse_com_command(self, ins):
        """Parse KEY, COM or STRIG syntax."""
        yield self._parse_bracket(ins)
        yield ins.require_read((tk.ON, tk.OFF, tk.STOP))

    def _parse_strig_switch(self, ins):
        """Parse STRIG ON/OFF syntax."""
        yield ins.require_read((tk.ON, tk.OFF))

    def _parse_on_event(self, ins):
        """Helper function for ON event trap definitions."""
        # token is known to be in (tk.PEN, tk.KEY, tk.TIMER, tk.PLAY, tk.COM, tk.STRIG)
        # before we call this generator
        token = ins.read_keyword_token()
        yield token
        if token != tk.PEN:
            yield self._parse_bracket(ins)
        else:
            yield None
        ins.require_read((tk.GOSUB,))
        yield self._parse_jumpnum(ins)
        ins.require_end()

    ###########################################################################
    # sound statements

    def _parse_beep(self, ins):
        """Parse BEEP syntax."""
        if self._syntax in ('pcjr', 'tandy'):
            # Tandy/PCjr BEEP ON, OFF
            yield ins.skip_blank_read_if((tk.ON, tk.OFF))
        else:
            yield None
        # if a syntax error happens, we still beeped.

    def _parse_noise(self, ins):
        """Parse NOISE syntax (Tandy/PCjr)."""
        yield self.parse_expression(ins)
        ins.require_read((b',',))
        yield self.parse_expression(ins)
        ins.require_read((b',',))
        yield self.parse_expression(ins)
        ins.require_end()

    def _parse_sound(self, ins):
        """Parse SOUND syntax."""
        command = None
        if self._syntax in ('pcjr', 'tandy'):
            # Tandy/PCjr SOUND ON, OFF
            command = ins.skip_blank_read_if((tk.ON, tk.OFF))
        if command:
            yield command
        else:
            yield self.parse_expression(ins)
            ins.require_read((b',',))
            dur = self.parse_expression(ins)
            yield dur
            # only look for args 3 and 4 if duration is > 0;
            # otherwise those args are a syntax error (on tandy)
            if (dur.sign() == 1) and ins.skip_blank_read_if((b',',)) and self._syntax in ('pcjr', 'tandy'):
                yield self.parse_expression(ins, allow_empty=True)
                if ins.skip_blank_read_if((b',',)):
                    yield self.parse_expression(ins)
                else:
                    yield None
            else:
                yield None
                yield None
        ins.require_end()

    def _parse_play(self, ins):
        """Parse PLAY (music) syntax."""
        if self._syntax in ('pcjr', 'tandy'):
            for _ in range(3):
                last = self.parse_expression(ins, allow_empty=True)
                yield last
                if not ins.skip_blank_read_if((b',',)):
                    break
            else:
                raise error.BASICError(error.STX)
            if last is None:
                raise error.BASICError(error.MISSING_OPERAND)
            ins.require_end()
        else:
            yield self.parse_expression(ins, allow_empty=True)
            ins.require_end(err=error.IFC)

    ###########################################################################
    # memory and machine port statements

    def _parse_def_seg(self, ins):
        """Parse DEF SEG syntax."""
        # must be uppercase in tokenised form, otherwise syntax error
        ins.require_read((tk.W_SEG,))
        if ins.skip_blank_read_if((tk.O_EQ,)):
            yield self.parse_expression(ins)
        else:
            yield None

    def _parse_def_usr(self, ins):
        """Parse DEF USR syntax."""
        ins.require_read((tk.USR,))
        yield ins.skip_blank_read_if(tk.DIGIT)
        ins.require_read((tk.O_EQ,))
        yield self.parse_expression(ins)

    def _parse_bload(self, ins):
        """Parse BLOAD syntax."""
        yield self.parse_expression(ins)
        if ins.skip_blank_read_if((b',',)):
            yield self.parse_expression(ins)
        else:
            yield None
        ins.require_end()

    def _parse_bsave(self, ins):
        """Parse BSAVE syntax."""
        yield self.parse_expression(ins)
        ins.require_read((b',',))
        yield self.parse_expression(ins)
        ins.require_read((b',',))
        yield self.parse_expression(ins)
        ins.require_end()

    def _parse_call(self, ins):
        """Parse CALL and CALLS syntax."""
        yield self.parse_name(ins)
        if ins.skip_blank_read_if((b'(',)):
            while True:
                yield self._parse_variable(ins)
                if not ins.skip_blank_read_if((b',',)):
                    break
            ins.require_read((b')',))
        ins.require_end()

    def _parse_wait(self, ins):
        """Parse WAIT syntax."""
        yield self.parse_expression(ins)
        ins.require_read((b',',))
        yield self.parse_expression(ins)
        if ins.skip_blank_read_if((b',',)):
            yield self.parse_expression(ins)
        else:
            yield None
        ins.require_end()

    def _parse_call_extension(self, ins):
        """Parse extension statement."""
        yield ins.read_name()
        while True:
            yield self.parse_expression(ins, allow_empty=True)
            if not ins.skip_blank_read_if((b',',)):
                break

    ###########################################################################
    # disk statements

    def _parse_name(self, ins):
        """Parse NAME syntax."""
        yield self.parse_expression(ins)
        # AS is not a tokenised word
        ins.require_read((tk.W_AS,))
        yield self.parse_expression(ins)

    ###########################################################################
    # clock statements

    def _parse_time_date(self, ins):
        """Parse TIME$ or DATE$ syntax."""
        ins.require_read((tk.O_EQ,))
        yield self.parse_expression(ins)
        ins.require_end()

    ##########################################################
    # code statements

    def _parse_delete_llist(self, ins):
        """Parse DELETE syntax."""
        yield self._parse_line_range(ins)
        ins.require_end()

    def _parse_edit(self, ins):
        """Parse EDIT syntax."""
        if ins.skip_blank() not in tk.END_STATEMENT:
            yield self._parse_jumpnum_or_dot(ins, err=error.IFC)
        else:
            yield None
        ins.require_end(err=error.IFC)

    def _parse_auto(self, ins):
        """Parse AUTO syntax."""
        yield self._parse_jumpnum_or_dot(ins, allow_empty=True)
        if ins.skip_blank_read_if((b',',)):
            inc = self._parse_optional_jumpnum(ins)
            if inc is None:
                raise error.BASICError(error.IFC)
            else:
                yield inc
        else:
            yield None
        ins.require_end()

    def _parse_save(self, ins):
        """Parse SAVE syntax."""
        yield self.parse_expression(ins)
        if ins.skip_blank_read_if((b',',)):
            yield ins.require_read((b'A', b'a', b'P', b'p'))
        else:
            yield None
        ins.require_end()

    def _parse_list(self, ins):
        """Parse LIST syntax."""
        yield self._parse_line_range(ins)
        if ins.skip_blank_read_if((b',',)):
            yield self.parse_expression(ins)
            # ignore everything after file spec
            ins.skip_to(tk.END_LINE)
        else:
            yield None
            ins.require_end()

    def _parse_load(self, ins):
        """Parse LOAD syntax."""
        yield self.parse_expression(ins)
        if ins.skip_blank_read_if((b',',)):
            yield ins.require_read((b'R', b'r'))
        else:
            yield None
        ins.require_end()

    def _parse_renum(self, ins):
        """Parse RENUM syntax."""
        new, old, step = None, None, None
        if ins.skip_blank() not in tk.END_STATEMENT:
            new = self._parse_jumpnum_or_dot(ins, allow_empty=True)
            if ins.skip_blank_read_if((b',',)):
                old = self._parse_jumpnum_or_dot(ins, allow_empty=True)
                if ins.skip_blank_read_if((b',',)):
                    # the number not optional at all if a comma is given
                    # but we need to control the error type
                    step = self._parse_optional_jumpnum(ins)
                    # negative numbers leave us before the end and raise STX
                    ins.require_end()
                    # empty afer comma raises IFC
                    if step is None:
                        raise error.BASICError(error.IFC)
        # we need require_end in both places to get correct error sequencing
        ins.require_end()
        for n in (new, old, step):
            yield n

    def _parse_chain(self, ins):
        """Parse CHAIN syntax."""
        yield ins.skip_blank_read_if((tk.MERGE,)) is not None
        yield self.parse_expression(ins)
        jumpnum, common_all, delete_range = None, False, True
        if ins.skip_blank_read_if((b',',)):
            # check for an expression that indicates a line in the other program.
            # This is not stored as a jumpnum (to avoid RENUM)
            jumpnum = self.parse_expression(ins, allow_empty=True)
            if ins.skip_blank_read_if((b',',)):
                common_all = ins.skip_blank_read_if((tk.W_ALL,), 3)
                if common_all:
                    # CHAIN "file", , ALL, DELETE
                    delete_range = ins.skip_blank_read_if((b',',))
                # CHAIN "file", , DELETE
        yield jumpnum
        yield common_all
        if delete_range and ins.skip_blank_read_if((tk.DELETE,)):
            from_line = self._parse_optional_jumpnum(ins)
            if ins.skip_blank_read_if((tk.O_MINUS,)):
                to_line = self._parse_optional_jumpnum(ins)
            else:
                to_line = from_line
            error.throw_if(not to_line)
            delete_lines = (from_line, to_line)
            # ignore rest if preceded by comma
            if ins.skip_blank_read_if((b',',)):
                ins.skip_to(tk.END_STATEMENT)
            yield delete_lines
        else:
            yield None
        ins.require_end()

    ###########################################################################
    # file statements

    def _parse_open(self, ins):
        """Parse OPEN syntax."""
        yield self.parse_expression(ins)
        first_syntax = ins.skip_blank_read_if((b',',))
        yield first_syntax
        if first_syntax:
            args = self._parse_open_first(ins)
        else:
            args = self._parse_open_second(ins)
        for a in args:
            yield a

    def _parse_open_first(self, ins):
        """Parse OPEN first ('old') syntax."""
        ins.skip_blank_read_if((b'#',))
        yield self.parse_expression(ins)
        ins.require_read((b',',))
        yield self.parse_expression(ins)
        if ins.skip_blank_read_if((b',',)):
            yield self.parse_expression(ins)
        else:
            yield None

    def _parse_open_second(self, ins):
        """Parse OPEN second ('new') syntax."""
        # mode clause
        if ins.skip_blank_read_if((tk.FOR,)):
            # read mode word
            if ins.skip_blank_read_if((tk.INPUT,)):
                yield b'I'
            else:
                mode_dict = {tk.W_OUTPUT: b'O', tk.W_RANDOM: b'R', tk.W_APPEND: b'A'}
                word = ins.skip_blank_read_if(mode_dict, 6)
                if word is not None:
                    yield mode_dict[word]
                else:
                    raise error.BASICError(error.STX)
        else:
            yield None
        # ACCESS clause
        if ins.skip_blank_read_if((tk.W_ACCESS,), 6):
            yield self._parse_read_write(ins)
        else:
            yield None
        # LOCK clause
        if ins.skip_blank_read_if((tk.LOCK,), 2):
            yield self._parse_read_write(ins)
        else:
            yield ins.skip_blank_read_if((tk.W_SHARED,), 6)
        # AS file number clause
        ins.require_read((tk.W_AS,))
        ins.skip_blank_read_if((b'#',))
        yield self.parse_expression(ins)
        # LEN clause
        if ins.skip_blank_read_if((tk.LEN,), 2):
            ins.require_read((tk.O_EQ,))
            yield self.parse_expression(ins)
        else:
            yield None

    def _parse_read_write(self, ins):
        """Parse access mode for OPEN."""
        d = ins.skip_blank_read_if((tk.READ, tk.WRITE))
        if d == tk.WRITE:
            return b'W'
        elif d == tk.READ:
            return b'RW' if ins.skip_blank_read_if((tk.WRITE,)) else b'R'
        raise error.BASICError(error.STX)

    def _parse_close(self, ins):
        """Parse CLOSE syntax."""
        if ins.skip_blank() not in tk.END_STATEMENT:
            while True:
                # if an error occurs, the files parsed before are closed anyway
                ins.skip_blank_read_if((b'#',))
                yield self.parse_expression(ins)
                if not ins.skip_blank_read_if((b',',)):
                    break

    def _parse_field(self, ins):
        """Parse FIELD syntax."""
        ins.skip_blank_read_if((b'#',))
        yield self.parse_expression(ins)
        if ins.skip_blank_read_if((b',',)):
            while True:
                yield self.parse_expression(ins)
                ins.require_read((tk.W_AS,), err=error.IFC)
                yield self._parse_variable(ins)
                if not ins.skip_blank_read_if((b',',)):
                    break

    def _parse_lock_unlock(self, ins):
        """Parse LOCK or UNLOCK syntax."""
        ins.skip_blank_read_if((b'#',))
        yield self.parse_expression(ins)
        if not ins.skip_blank_read_if((b',',)):
            ins.require_end()
            yield None
            yield None
        else:
            expr = self.parse_expression(ins, allow_empty=True)
            yield expr
            if ins.skip_blank_read_if((tk.TO,)):
                yield self.parse_expression(ins)
            elif expr is not None:
                yield None
            else:
                raise error.BASICError(error.MISSING_OPERAND)

    def _parse_ioctl(self, ins):
        """Parse IOCTL syntax."""
        ins.skip_blank_read_if((b'#',))
        yield self.parse_expression(ins)
        ins.require_read((b',',))
        yield self.parse_expression(ins)

    def _parse_put_get_file(self, ins):
        """Parse PUT and GET syntax."""
        ins.skip_blank_read_if((b'#',))
        yield self.parse_expression(ins)
        if ins.skip_blank_read_if((b',',)):
            yield self.parse_expression(ins)
        else:
            yield None

    ###########################################################################
    # graphics statements

    def _parse_pair(self, ins):
        """Parse coordinate pair."""
        ins.require_read((b'(',))
        yield self.parse_expression(ins)
        ins.require_read((b',',))
        yield self.parse_expression(ins)
        ins.require_read((b')',))

    def _parse_pset_preset(self, ins):
        """Parse PSET and PRESET syntax."""
        yield ins.skip_blank_read_if((tk.STEP,))
        for c in self._parse_pair(ins):
            yield c
        if ins.skip_blank_read_if((b',',)):
            yield self.parse_expression(ins)
        else:
            yield None
        ins.require_end()

    def _parse_window(self, ins):
        """Parse WINDOW syntax."""
        screen = ins.skip_blank_read_if((tk.SCREEN,))
        yield screen
        if ins.skip_blank() == b'(':
            for c in self._parse_pair(ins):
                yield c
            ins.require_read((tk.O_MINUS,))
            for c in self._parse_pair(ins):
                yield c
        elif screen:
            raise error.BASICError(error.STX)

    def _parse_circle(self, ins):
        """Parse CIRCLE syntax."""
        yield ins.skip_blank_read_if((tk.STEP,))
        for c in self._parse_pair(ins):
            yield c
        ins.require_read((b',',))
        last = self.parse_expression(ins)
        yield last
        for count_args in range(4):
            if ins.skip_blank_read_if((b',',)):
                last = self.parse_expression(ins, allow_empty=True)
                yield last
            else:
                break
        if last is None:
            raise error.BASICError(error.MISSING_OPERAND)
        for _ in range(count_args, 4):
            yield None
        ins.require_end()

    def _parse_paint(self, ins):
        """Parse PAINT syntax."""
        yield ins.skip_blank_read_if((tk.STEP,))
        for last in self._parse_pair(ins):
            yield last
        for count_args in range(3):
            if ins.skip_blank_read_if((b',',)):
                last = self.parse_expression(ins, allow_empty=True)
                yield last
            else:
                break
        if last is None:
            raise error.BASICError(error.MISSING_OPERAND)
        for _ in range(count_args, 3):
            yield None

    def _parse_view(self, ins):
        """Parse VIEW syntax."""
        yield ins.skip_blank_read_if((tk.SCREEN,))
        if ins.skip_blank() == b'(':
            for c in self._parse_pair(ins):
                yield c
            ins.require_read((tk.O_MINUS,))
            for c in self._parse_pair(ins):
                yield c
            if ins.skip_blank_read_if((b',',)):
                fill_comma = True
                fill = self.parse_expression(ins, allow_empty=True)
                yield fill
            else:
                fill_comma = False
                yield None
            if ins.skip_blank_read_if((b',',)):
                yield self.parse_expression(ins)
            else:
                error.throw_if(fill_comma and not fill, error.MISSING_OPERAND)
                yield None

    def _parse_line(self, ins):
        """Parse LINE syntax."""
        if ins.skip_blank() in (b'(', tk.STEP):
            yield ins.skip_blank_read_if((tk.STEP,))
            for c in self._parse_pair(ins):
                yield c
        else:
            for _ in range(3):
                yield None
        ins.require_read((tk.O_MINUS,))
        yield ins.skip_blank_read_if((tk.STEP,))
        for c in self._parse_pair(ins):
            yield c
        if ins.skip_blank_read_if((b',',)):
            expr = self.parse_expression(ins, allow_empty=True)
            yield expr
            if ins.skip_blank_read_if((b',',)):
                if ins.skip_blank_read_if((b'B',)):
                    shape = b'BF' if ins.skip_blank_read_if((b'F',)) else b'B'
                else:
                    shape = None
                yield shape
                if ins.skip_blank_read_if((b',',)):
                    yield self.parse_expression(ins)
                else:
                    # mustn't end on a comma
                    # mode == '' if nothing after previous comma
                    error.throw_if(not shape, error.STX)
                    yield None
            elif not expr:
                raise error.BASICError(error.MISSING_OPERAND)
            else:
                yield None
                yield None
        else:
            yield None
            yield None
            yield None
        ins.require_end()

    def _parse_get_graph(self, ins):
        """Parse graphics GET syntax."""
        # don't accept STEP for first coord
        for c in self._parse_pair(ins):
            yield c
        ins.require_read((tk.O_MINUS,))
        yield ins.skip_blank_read_if((tk.STEP,))
        for c in self._parse_pair(ins):
            yield c
        ins.require_read((b',',))
        yield self.parse_name(ins)
        ins.require_end()

    def _parse_put_graph(self, ins):
        """Parse graphics PUT syntax."""
        # don't accept STEP
        for c in self._parse_pair(ins):
            yield c
        ins.require_read((b',',))
        yield self.parse_name(ins)
        if ins.skip_blank_read_if((b',',)):
            yield ins.require_read((tk.PSET, tk.PRESET, tk.AND, tk.OR, tk.XOR))
        else:
            yield None
        ins.require_end()

    ###########################################################################
    # variable statements

    def _parse_clear(self, ins):
        """Parse CLEAR syntax."""
        # integer expression allowed but ignored
        yield self.parse_expression(ins, allow_empty=True)
        if ins.skip_blank_read_if((b',',)):
            exp1 = self.parse_expression(ins, allow_empty=True)
            yield exp1
            if not ins.skip_blank_read_if((b',',)):
                if not exp1:
                    raise error.BASICError(error.STX)
            else:
                # set aside stack space for GW-BASIC. The default is the previous stack space size.
                exp2 = self.parse_expression(ins, allow_empty=True)
                yield exp2
                if self._syntax in ('pcjr', 'tandy') and ins.skip_blank_read_if((b',',)):
                    # Tandy/PCjr: select video memory size
                    yield self.parse_expression(ins)
                elif not exp2:
                    raise error.BASICError(error.STX)
        ins.require_end()

    def _parse_def_fn(self, ins):
        """DEF FN: define a function."""
        ins.require_read((tk.FN,))
        yield self.parse_name(ins)

    def _parse_var_list(self, ins):
        """Generator: lazily parse variable list."""
        while True:
            yield self._parse_variable(ins)
            if not ins.skip_blank_read_if((b',',)):
                break

    def _parse_deftype(self, ins):
        """Parse DEFSTR/DEFINT/DEFSNG/DEFDBL syntax."""
        while True:
            start = ins.require_read(tuple(iterchar(LETTERS)))
            stop = None
            if ins.skip_blank_read_if((tk.O_MINUS,)):
                stop = ins.require_read(tuple(iterchar(LETTERS)))
            yield start, stop
            if not ins.skip_blank_read_if((b',',)):
                break

    def _parse_erase(self, ins):
        """Parse ERASE syntax."""
        while True:
            yield self.parse_name(ins)
            if not ins.skip_blank_read_if((b',',)):
                break

    def _parse_let(self, ins):
        """Parse LET, LSET or RSET syntax."""
        yield self._parse_variable(ins)
        ins.require_read((tk.O_EQ,))
        # we're not using a temp string here
        # as it would delete the new string generated by let if applied to a code literal
        yield self.parse_expression(ins)

    def _parse_mid(self, ins):
        """Parse MID$ syntax."""
        # do not use require_read as we don't allow whitespace here
        if ins.read(1) != b'(':
            raise error.BASICError(error.STX)
        yield self._parse_variable(ins)
        ins.require_read((b',',))
        yield self.parse_expression(ins)
        if ins.skip_blank_read_if((b',',)):
            yield self.parse_expression(ins)
        else:
            yield None
        ins.require_read((b')',))
        ins.require_read((tk.O_EQ,))
        # we're not using a temp string here
        # as it would delete the new string generated by midset if applied to a code literal
        yield self.parse_expression(ins)
        ins.require_end()

    def _parse_option_base(self, ins):
        """Parse OPTION BASE syntax."""
        ins.require_read((tk.W_BASE,))
        # MUST be followed by ASCII '1' or '0', num constants or expressions are an error!
        yield ins.require_read((b'0', b'1'))

    def _parse_prompt(self, ins):
        """Helper function for INPUT: parse prompt definition."""
        # ; to avoid echoing newline
        newline = not ins.skip_blank_read_if((b';',))
        # parse prompt
        prompt, following = b'', b';'
        if ins.skip_blank() == b'"':
            # only literal allowed, not a string expression
            prompt = ins.read_string().strip(b'"')
            following = ins.require_read((b';', b','))
        return newline, prompt, following

    def _parse_input(self, ins):
        """Parse INPUT syntax."""
        if ins.skip_blank_read_if((b'#',)):
            yield self.parse_expression(ins)
            ins.require_read((b',',))
        else:
            yield None
            yield self._parse_prompt(ins)
        for arg in self._parse_var_list(ins):
            yield arg

    def _parse_line_input(self, ins):
        """Parse LINE INPUT syntax."""
        ins.require_read((tk.INPUT,))
        if ins.skip_blank_read_if((b'#',)):
            yield self.parse_expression(ins)
            ins.require_read((b',',))
        else:
            yield None
            yield self._parse_prompt(ins)
        # get string variable
        yield self._parse_variable(ins)

    def _parse_restore(self, ins):
        """Parse RESTORE syntax."""
        if ins.skip_blank() == tk.T_UINT:
            yield self._parse_jumpnum(ins)
            ins.require_end()
        else:
            # undefined line number for syntax errors if no line number given
            ins.require_end(err=error.UNDEFINED_LINE_NUMBER)
            yield None

    def _parse_swap(self, ins):
        """Parse SWAP syntax."""
        yield self._parse_variable(ins)
        ins.require_read((b',',))
        yield self._parse_variable(ins)

    ###########################################################################
    # console / text screen statements

    def _parse_key_macro(self, ins):
        """Parse KEY ON/OFF/LIST syntax."""
        yield ins.read_keyword_token()

    def _parse_cls(self, ins):
        """Parse CLS syntax."""
        if self._syntax != 'pcjr':
            yield self.parse_expression(ins, allow_empty=True)
            # optional comma
            if not ins.skip_blank_read_if((b',',)):
                ins.require_end(err=error.IFC)
        else:
            yield None

    def _parse_color(self, ins):
        """Parse COLOR syntax."""
        last = self.parse_expression(ins, allow_empty=True)
        yield last
        if ins.skip_blank_read_if((b',',)):
            # unlike LOCATE, ending in any number of commas is a Missing Operand
            while True:
                last = self.parse_expression(ins, allow_empty=True)
                yield last
                if not ins.skip_blank_read_if((b',',)):
                    break
            if last is None:
                raise error.BASICError(error.MISSING_OPERAND)
        elif last is None:
            raise error.BASICError(error.IFC)

    def _parse_palette(self, ins):
        """Parse PALETTE syntax."""
        attrib = self.parse_expression(ins, allow_empty=True)
        yield attrib
        if attrib is None:
            yield None
            ins.require_end()
        else:
            ins.require_read((b',',))
            colour = self.parse_expression(ins, allow_empty=True)
            yield colour
            error.throw_if(attrib is None or colour is None, error.STX)

    def _parse_palette_using(self, ins):
        """Parse PALETTE USING syntax."""
        ins.require_read((tk.USING,))
        array_name, start_indices = self._parse_variable(ins)
        yield array_name, start_indices
        # brackets are not optional
        error.throw_if(not start_indices, error.STX)

    def _parse_locate(self, ins):
        """Parse LOCATE syntax."""
        #row, col, cursor, start, stop
        for i in range(5):
            yield self.parse_expression(ins, allow_empty=True)
            # note that LOCATE can end on a 5th comma but no stuff allowed after it
            if not ins.skip_blank_read_if((b',',)):
                break
        ins.require_end()

    def _parse_view_print(self, ins):
        """Parse VIEW PRINT syntax."""
        ins.require_read((tk.PRINT,))
        start = self.parse_expression(ins, allow_empty=True)
        yield start
        if start is not None:
            ins.require_read((tk.TO,))
            yield self.parse_expression(ins)
        else:
            yield None
        ins.require_end()

    def _parse_write(self, ins):
        """Parse WRITE syntax."""
        if ins.skip_blank_read_if((b'#',)):
            yield self.parse_expression(ins)
            ins.require_read((b',',))
        else:
            yield None
        if ins.skip_blank() not in tk.END_STATEMENT:
            while True:
                yield self.parse_expression(ins)
                if not ins.skip_blank_read_if((b',', b';')):
                    break
            ins.require_end()

    def _parse_width(self, ins):
        """Parse WIDTH syntax."""
        d = ins.skip_blank_read_if((b'#', tk.LPRINT))
        if d:
            if d == b'#':
                yield self.parse_expression(ins)
                ins.require_read((b',',))
            else:
                yield tk.LPRINT
            yield self.parse_expression(ins)
        else:
            yield None
            if ins.peek() in set(iterchar(DIGITS)) | set(tk.NUMBER):
                expr = self.expression_parser.read_number_literal(ins)
            else:
                expr = self.parse_expression(ins)
            yield expr
            if isinstance(expr, values.String):
                ins.require_read((b',',))
                yield self.parse_expression(ins)
            elif not ins.skip_blank_read_if((b',',)):
                yield None
                ins.require_end(error.IFC)
            else:
                # parse dummy number rows setting
                yield self.parse_expression(ins, allow_empty=True)
                # trailing comma is accepted
                ins.skip_blank_read_if((b',',))
        ins.require_end()

    def _parse_screen(self, ins):
        """Parse SCREEN syntax."""
        # erase can only be set on pcjr/tandy 5-argument syntax
        # all but last arguments are optional and may be followed by a comma
        argcount = 0
        while True:
            last = self.parse_expression(ins, allow_empty=True)
            yield last
            argcount += 1
            if not ins.skip_blank_read_if((b',',)):
                break
        if last is None:
            if self._syntax == 'tandy' and argcount == 1:
                raise error.BASICError(error.IFC)
            raise error.BASICError(error.MISSING_OPERAND)
        for _ in range(argcount, 5):
            yield None
        ins.require_end()

    def _parse_pcopy(self, ins):
        """Parse PCOPY syntax."""
        yield self.parse_expression(ins)
        ins.require_read((b',',))
        yield self.parse_expression(ins)
        ins.require_end()

    def _parse_print(self, ins, parse_file):
        """Parse PRINT or LPRINT syntax."""
        if parse_file:
            if ins.skip_blank_read_if((b'#',)):
                yield self.parse_expression(ins)
                ins.require_read((b',',))
            else:
                yield None
        while True:
            d = ins.skip_blank_read()
            if d in tk.END_STATEMENT:
                ins.seek(-len(d), 1)
                break
            elif d == tk.USING:
                yield (tk.USING, None)
                yield self.parse_expression(ins)
                ins.require_read((b';',))
                has_args = False
                while True:
                    expr = self.parse_expression(ins, allow_empty=True)
                    yield expr
                    if expr is None:
                        ins.require_end()
                        # need at least one argument after format string
                        if not has_args:
                            raise error.BASICError(error.MISSING_OPERAND)
                        break
                    has_args = True
                    if not ins.skip_blank_read_if((b';', b',')):
                        break
                break
            elif d in (b',', b';'):
                yield (d, None)
            elif d in (tk.SPC, tk.TAB):
                num = self.parse_expression(ins)
                ins.require_read((b')',))
                yield (d, num)
            else:
                ins.seek(-len(d), 1)
                yield (None, None)
                yield self.parse_expression(ins)

    ###########################################################################
    # loops and branches

    def _parse_on_jump(self, ins):
        """ON: calculated jump."""
        yield self.parse_expression(ins)
        yield ins.require_read((tk.GOTO, tk.GOSUB))
        while True:
            num = self._parse_optional_jumpnum(ins)
            yield num
            if not ins.skip_blank_read_if((b',',)):
                break
        ins.require_end()

    def _parse_if(self, ins):
        """IF: enter branching statement."""
        # avoid overflow: don't use bools.
        condition = self.parse_expression(ins)
        # optional comma
        ins.skip_blank_read_if((b',',))
        ins.require_read((tk.THEN, tk.GOTO))
        # THEN and GOTO tokens both have length 1
        start_pos = ins.tell() - 1
        # allow cofunction to evaluate condition
        branch = yield condition
        # we only even parse the ELSE clause if this is false
        if branch:
            jumpnum = self._parse_optional_jumpnum(ins)
            yield jumpnum
            if jumpnum is None:
                ins.seek(start_pos)
        else:
            # find correct ELSE block, if any
            # ELSEs may be nested in the THEN clause
            nesting_level = 0
            while True:
                d = ins.skip_to_read(tk.END_STATEMENT + (tk.IF,))
                if d == tk.IF:
                    # nesting step on IF. (it's less convenient to count THENs
                    # because they could be THEN or GOTO)
                    nesting_level += 1
                elif d == b':':
                    # :ELSE is ELSE; may be whitespace in between. no : means it's ignored.
                    if ins.skip_blank_read_if((tk.ELSE,)):
                        # ELSE has length 1
                        start_pos = ins.tell() - 1
                        if nesting_level > 0:
                            nesting_level -= 1
                        else:
                            jumpnum = self._parse_optional_jumpnum(ins)
                            yield jumpnum
                            if jumpnum is None:
                                ins.seek(start_pos)
                            break
                else:
                    # end of line, don't look for line number
                    ins.seek(-len(d), 1)
                    yield None
                    break

    def _parse_for(self, ins):
        """Parse FOR syntax."""
        # read variable
        yield self.parse_name(ins)
        ins.require_read((tk.O_EQ,))
        yield self.parse_expression(ins)
        ins.require_read((tk.TO,))
        yield self.parse_expression(ins)
        if ins.skip_blank_read_if((tk.STEP,)):
            yield self.parse_expression(ins)
        else:
            yield None
        ins.require_end()

    def _parse_next(self, ins):
        """Parse NEXT syntax."""
        # note that next_ will not run the full generator if it finds a loop to iterate
        while True:
            # optional var name, errors have been checked during _find_next scan
            if ins.skip_blank() not in tk.END_STATEMENT + (b',',):
                yield self.parse_name(ins)
            else:
                yield None
            # done if we're not jumping into a comma'ed NEXT
            if not ins.skip_blank_read_if((b',')):
                break



"""
PC-BASIC - GW-BASIC/BASICA/Cartridge BASIC compatible interpreter

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

# compatibility pre-init: ensures package __path__ is absolute
from . import compat

from .basic import __version__
from .basic import NAME, VERSION, AUTHOR, COPYRIGHT
from .basic import Session, codepage, font
from .main import main, script_entry_point_guard



"""
PC-BASIC - clipboard.py
Clipboard handling

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import subprocess
import locale

from ..basic.base import signals
from ..basic.base import scancode
from ..compat import which

ENCODING = locale.getpreferredencoding()


class Clipboard(object):
    """Clipboard handling interface."""

    def __init__(self):
        """Initialise the clipboard handler."""
        self.ok = True

    def copy(self, text):
        """Put unicode text on clipboard."""
        pass

    def paste(self):
        """Return unicode text from clipboard."""
        return u''


class MacClipboard(Clipboard):
    """Clipboard handling for OSX. Deprecated, used only by PyGame interface."""

    def paste(self):
        """Get unicode text from clipboard."""
        return (
            subprocess.check_output('pbpaste').decode(ENCODING, 'replace')
            .replace(u'\r\n', u'\r').replace(u'\n', u'\r')
        )

    def copy(self, text):
        """Put unicode text on clipboard."""
        try:
            p = subprocess.Popen('pbcopy', stdin=subprocess.PIPE)
            p.communicate(text.encode(ENCODING, 'replace'))
        except subprocess.CalledProcessError:
            pass


class XClipboard(Clipboard):
    """Clipboard handling for X Window System using xsel or xclip."""

    def __init__(self):
        """Check for presence of xsel or xclip."""
        if which('xclip'):
            self._command = 'xclip'
            self._notmouse = ['-selection', 'clipboard']
            self.ok = True
        elif which('xsel'):
            # note that xsl has a bug that makes chromium/atom hang on paste
            # https://github.com/electron/electron/issues/3116
            self._command = 'xsel'
            self._notmouse = ['-b']
            self.ok = True
        else:
            self.ok = False

    def paste(self):
        """Get unicode text from clipboard."""
        output = subprocess.check_output([self._command, '-o'] + self._notmouse)
        return (output.decode(ENCODING, 'replace').replace(u'\r\n', u'\r').replace(u'\n', u'\r'))

    def copy(self, text):
        """Put unicode text on clipboard."""
        try:
            p = subprocess.Popen([self._command, '-i'] + self._notmouse, stdin=subprocess.PIPE)
            p.communicate(text.encode(ENCODING, 'replace'))
        except subprocess.CalledProcessError:
            pass


##############################################################################

class ClipboardInterface(object):
    """Clipboard user interface."""

    def __init__(
            self, clipboard_handler, input_queue,
            width, height, font_width, font_height, size):
        """Initialise clipboard feedback handler."""
        self._input_queue = input_queue
        self._active = False
        self.select_start = None
        self.select_stop = None
        self.selection_rect = None
        self.width = width
        self.height = height
        self.font_width = font_width
        self.font_height = font_height
        self.size = size
        self._clipboard_handler = clipboard_handler

    def active(self):
        """True if clipboard mode is active."""
        return self._active

    def start(self, r, c):
        """Enter clipboard mode (clipboard key pressed)."""
        self._active = True
        if c < 1:
            r -= 1
            c = self.width
        if c > self.width:
            r += 1
            c = 1
        if r > self.height:
            r, c = self.height, self.width
        if r < 1:
            r, c = 1, 1
        self.select_start = [r, c]
        self.select_stop = [r, c]
        self.selection_rect = []

    def stop(self):
        """Leave clipboard mode (clipboard key released)."""
        self._active = False
        self.select_start = None
        self.select_stop = None
        self.selection_rect = None

    def copy(self):
        """Copy screen characters from selection into clipboard."""
        start, stop = self.select_start, self.select_stop
        if not start or not stop:
            return
        if start[0] == stop[0] and start[1] == stop[1]:
            return
        if start[0] > stop[0] or (start[0] == stop[0] and start[1] > stop[1]):
            start, stop = stop, start
        self._input_queue.put(signals.Event(
                signals.CLIP_COPY, (start[0], start[1], stop[0], stop[1])))

    def paste(self, text):
        """Paste from clipboard into keyboard buffer."""
        self._input_queue.put(signals.Event(signals.CLIP_PASTE, (text,)))

    def move(self, r, c):
        """Move the head of the selection and update feedback."""
        self.select_stop = [r, c]
        start, stop = self.select_start, self.select_stop
        if stop[1] < 1:
            stop[0] -= 1
            stop[1] = self.width+1
        if stop[1] > self.width+1:
            stop[0] += 1
            stop[1] = 1
        if stop[0] > self.height:
            stop[:] = [self.height, self.width+1]
        if stop[0] < 1:
            stop[:] = [1, 1]
        if start[0] > stop[0] or (start[0] == stop[0] and start[1] > stop[1]):
            start, stop = stop, start
        rect_left = (start[1] - 1) * self.font_width
        rect_top = (start[0] - 1) * self.font_height
        rect_right = (stop[1] - 1) * self.font_width
        rect_bot = stop[0] * self.font_height
        if start[0] == stop[0]:
            # single row selection
            self.selection_rect = [(rect_left, rect_top, rect_right-rect_left, rect_bot-rect_top)]
        else:
            # multi-row selection
            self.selection_rect = [
                (rect_left, rect_top, self.size[0]-rect_left, self.font_height),
                (0, rect_top+self.font_height, self.size[0], rect_bot-rect_top-2*self.font_height),
                (0, rect_bot-self.font_height, rect_right, self.font_height)
            ]

    def handle_key(self, scan, c):
        """Handle keyboard clipboard commands."""
        if not self._active:
            return
        if c.upper() == u'C':
            self.copy()
        elif c.upper() == u'V':
            text = self._clipboard_handler.paste()
            self.paste(text)
        elif c.upper() == u'A':
            # select all
            self.select_start = [1, 1]
            self.move(self.height, self.width+1)
        elif scan == scancode.LEFT:
            # move selection head left
            self.move(self.select_stop[0], self.select_stop[1]-1)
        elif scan == scancode.RIGHT:
            # move selection head right
            self.move(self.select_stop[0], self.select_stop[1]+1)
        elif scan == scancode.UP:
            # move selection head up
            self.move(self.select_stop[0]-1, self.select_stop[1])
        elif scan == scancode.DOWN:
            # move selection head down
            self.move(self.select_stop[0]+1, self.select_stop[1])



"""
PC-BASIC - data.fonts
Font files

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""


import sys
import logging
import binascii
import unicodedata

from ...compat import resources, iteritems, itervalues, unichr, iterchar


_HEIGHTS = (8, 14, 16)
_DEFAULT_NAME = 'default'
_FONT_PATTERN = '{name}_{height:02d}.hex'

# (deprecated) aliases for the default font. should not include 'default'
_DEFAULT_ALIASES = ('freedos', 'univga', 'unifont')


# list of available fonts - top level names only
FONTS = tuple(sorted(set(
    name.split(u'_', 1)[0]
    for name in resources.contents(__package__)
    if name.lower().endswith('.hex') and resources.is_resource(__package__, name)
)))


def read_fonts(codepage_dict, font_families):
    """Load font typefaces."""
    # default font is fallback
    font_families = (_DEFAULT_NAME,) + tuple(font_families)
    # load the graphics fonts, including the 8-pixel RAM font
    # use set() for speed - lookup is O(1) rather than O(n) for list
    unicode_needed = set(itervalues(codepage_dict))
    # break up any grapheme clusters and add components to set of needed glyphs
    unicode_needed |= set(c for cluster in unicode_needed if len(cluster) > 1 for c in cluster)
    # load font resources
    font_files = {
        _height: [
            _font for _font in
            (_read_font_file(_name, _height) for _name in font_families)
            if _font is not None
        ]
        for _height in _HEIGHTS
    }
    # convert
    uc_fonts = {
        _height: load_hex(_font_file, _height, unicode_needed)
        for _height, _font_file in iteritems(font_files)
        if _font_file
    }
    return uc_fonts


def _read_font_file(name, height):
    """Get contents of font file."""
    fontname = _FONT_PATTERN.format(name=name, height=height)
    try:
        return resources.read_binary(__package__, fontname)
    except EnvironmentError as e:
        if name in _DEFAULT_ALIASES:
            return _read_font_file(_DEFAULT_NAME, height)
        else:
            logging.debug('Failed to load %d-pixel font `%s`: %s', height, name, e)


###################################################################################################
# hex font loader

def load_hex(hex_resources, height, all_needed):
    """Load a set of overlaying unifont .hex files."""
    fontdict = {}
    # get elements of codepoint sequences in case we need to combine glyphs
    elements = set.union(*(set(unicodedata.normalize('NFD', _c)) for _c in all_needed))
    all_needed = set(all_needed).union(elements)
    # transform the (smaller) set of needed chars into sequences for comparison
    # rather than the (larger) set of available sequences into chars
    needed_sequences = set(b','.join(b'%04X' % ord(_c) for _c in _s) for _s in all_needed)
    missing = set(all_needed)
    for hexres in reversed(hex_resources):
        if hexres is None:
            continue
        missing = _get_glyphs_from_hex(height, hexres, fontdict, missing, needed_sequences)
    missing = _combine_glyphs(height, fontdict, missing)
    # fill missing with nulls
    fontdict.update({_u: b'\0' * height for _u in missing})
    # char 0 should always be defined and empty
    fontdict[u'\0'] = b'\0' * height
    missing -= {u'\0'}
    # warn if we miss needed glyphs
    _warn_missing(height, missing)
    return fontdict


def _get_glyphs_from_hex(height, hexres, fontdict, missing, needed_sequences):
    for line in hexres.splitlines():
        # ignore empty lines and comment lines (first char is #)
        if (not line) or (line[:1] == b'#'):
            continue
        # strip off comments
        # split unicodepoint and hex string (max 32 chars)
        ucs_str, fonthex = line.split(b':')
        # get rid of spaces
        ucs_str = b''.join(ucs_str.split()).upper()
        # skip grapheme clusters we won't need
        if ucs_str not in needed_sequences:
            continue
        # remove from needed list
        needed_sequences -= {ucs_str}
        ucs_sequence = ucs_str.split(b',')
        fonthex = fonthex.split(b'#')[0].strip()
        # extract codepoint and hex string;
        # discard anything following whitespace; ignore malformed lines
        try:
            # construct grapheme cluster
            c = u''.join(unichr(int(_ucshex.strip(), 16)) for _ucshex in ucs_sequence)
            # skip chars we already have
            if (c in fontdict):
                continue
            # cut to required font size
            if len(fonthex) < 64:
                # 8xN glyph
                fonthex = fonthex[:2*height]
            else:
                # 16x16 glyph
                fonthex = fonthex[:4*height]
            fontdict[c] = binascii.unhexlify(fonthex)
        except Exception as e:
            logging.warning('Could not parse line in font file: %s: %s', repr(line), e)
        # remove newly found char
        # stop if we have all we need
        missing -= {c}
        if not missing:
            break
    return missing


def _combine_glyphs(height, fontdict, missing):
    """Fix missing grapheme clusters by combining components."""
    success = set()
    for cluster in missing:
        if cluster not in fontdict:
            # fully decompose the grapheme cluster
            decomposed = unicodedata.normalize('NFD', cluster)
            # try to combine grapheme clusters first
            if len(decomposed) > 1:
                # combine strings
                clusterglyph = bytearray(height)
                try:
                    for c in decomposed:
                        for y, row in enumerate(iterchar(fontdict[c])):
                            clusterglyph[y] |= ord(row)
                except KeyError as e:
                    logging.debug(
                        '%d-pixel font: Could not combine grapheme cluster %s, missing u+%04x [%s]',
                        height, cluster, ord(c), c
                    )
                fontdict[cluster] = bytes(clusterglyph)
                success.add(cluster)
    return missing - success


def _warn_missing(height, missing, max_warnings=5):
    """Warn if we miss needed glyphs."""
    warnings = 0
    for u in missing:
        warnings += 1
        sequence = ','.join('u+%04x' % ord(_c) for _c in u)
        logging.debug('Code point sequence %s not represented in %d-pixel font', sequence, height)
        if warnings == max_warnings:
            logging.debug('Further code point warnings suppressed.')
            break
